##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Test Suite
**********

:module: test_suite

:synopsis: Create a generic test-suite based on the connected modules, and
    gray test-suite for Message Protocol / TestApp usage.

.. currentmodule:: generic_test_suite


"""
from __future__ import annotations

import logging
import unittest
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Dict, List, Union
from unittest.suite import _isnotsuite

from .. import message
from ..auxiliary import AuxiliaryInterface
from .test_message_handler import test_app_interaction

if TYPE_CHECKING:
    from ..test_result.text_result import BannerTestResult
    from .test_case import BasicTest

__all__ = [
    "BaseTestSuite",
    "BasicTestSuiteSetup",
    "BasicTestSuiteTeardown",
    "RemoteTestSuiteSetup",
    "RemoteTestSuiteTeardown",
    "BasicTestSuite",
]

log = logging.getLogger(__name__)


class BaseTestSuite(unittest.TestCase):
    response_timeout = 10

    def __init__(
        self,
        test_suite_id: int,
        test_case_id: int,
        aux_list: Union[List[AuxiliaryInterface], None],
        setup_timeout: Union[int, None],
        run_timeout: Union[int, None],
        teardown_timeout: Union[int, None],
        test_ids: Union[dict, None],
        tag: Union[Dict[str, List[str]], None],
        *args: Any,
        **kwargs: Any,
    ):
        """Initialize generic test-case.

        :param test_suite_id: test suite identification number
        :param test_case_id: test case identification number
        :param aux_list: list of used auxiliaries
        :param setup_timeout: maximum time (in seconds) used to wait
            for a report during setup execution
        :param run_timeout: maximum time (in seconds) used to wait for
            a report during test_run execution
        :param teardown_timeout: the maximum time (in seconds) used to
            wait for a report during teardown execution
        :param test_ids: jama references to get the coverage
            eg: {"Component1": ["Req1", "Req2"], "Component2": ["Req3"]}
        :param tag: dictionary containing lists of variants and/or test levels
            when only a subset of tests needs to be executed
        """
        # Initialize base class
        super().__init__(*args, **kwargs)
        # Save test information
        self.test_auxiliary_list = aux_list or []
        self.test_suite_id = test_suite_id
        self.test_case_id = test_case_id
        self.test_ids = test_ids
        self.tag = tag
        self.start_time = self.stop_time = self.elapsed_time = 0

    def cleanup_and_skip(self, aux: AuxiliaryInterface, info_to_print: str):
        """Cleanup auxiliary and log reasons.

        :param aux: corresponding auxiliary to abort
        :param info_to_print: A message you want to print while cleaning up the test
        """
        # Log error message
        log.critical(info_to_print)

        # Send aborts to corresponding auxiliary
        if aux.send_abort_command(timeout=10) is not True:
            log.critical(f"Error occurred during abort command on auxiliary {aux}")

        self.fail(info_to_print)

    def __str__(self):
        return "%s (%s)" % (self._testMethodName, self.__class__.__name__)


class BasicTestSuiteSetup(BaseTestSuite):
    """Inherit from unittest testCase and represent setup fixture."""

    def __init__(
        self,
        test_suite_id: int,
        test_case_id: int,
        aux_list: Union[List[AuxiliaryInterface], None],
        setup_timeout: Union[int, None],
        run_timeout: Union[int, None],
        teardown_timeout: Union[int, None],
        test_ids: Union[dict, None],
        tag: Union[Dict[str, List[str]], None],
        *args: Any,
        **kwargs: Any,
    ):
        """Initialize Message Protocol / TestApp test-case.

        :param test_suite_id: test suite identification number
        :param test_case_id: test case identification number
        :param aux_list: list of used auxiliaries
        :param setup_timeout: maximum time (in seconds) used to wait
            for a report during setup execution
        :param run_timeout: maximum time (in seconds) used to wait for
            a report during test_run execution
        :param teardown_timeout: the maximum time (in seconds) used to
            wait for a report during teardown execution
        :param test_ids: jama references to get the coverage
            eg: {"Component1": ["Req1", "Req2"], "Component2": ["Req3"]}
        :param tag: dictionary containing lists of variants and/or test levels
            when only a subset of tests needs to be executed
        """
        super().__init__(
            test_suite_id,
            test_case_id,
            aux_list,
            setup_timeout,
            run_timeout,
            teardown_timeout,
            test_ids,
            tag,
            *args,
            **kwargs,
        )
        if any([setup_timeout, run_timeout, teardown_timeout]):
            log.warning("BasicTestSuiteSetup does not support test timeouts, it will be discarded")

    def test_suite_setUp(self):
        """Test method for constructing the actual test suite."""
        pass


class BasicTestSuiteTeardown(BaseTestSuite):
    """Inherit from unittest testCase and represent teardown fixture."""

    def __init__(
        self,
        test_suite_id: int,
        test_case_id: int,
        aux_list: Union[List[AuxiliaryInterface], None],
        setup_timeout: Union[int, None],
        run_timeout: Union[int, None],
        teardown_timeout: Union[int, None],
        test_ids: Union[dict, None],
        tag: Union[Dict[str, List[str]], None],
        *args: Any,
        **kwargs: Any,
    ):
        """Initialize Message Protocol / TestApp test-case.

        :param test_suite_id: test suite identification number
        :param test_case_id: test case identification number
        :param aux_list: list of used auxiliaries
        :param setup_timeout: maximum time (in seconds) used to wait
            for a report during setup execution
        :param run_timeout: maximum time (in seconds) used to wait for
            a report during test_run execution
        :param teardown_timeout: the maximum time (in seconds) used to
            wait for a report during teardown execution
        :param test_ids: jama references to get the coverage
            eg: {"Component1": ["Req1", "Req2"], "Component2": ["Req3"]}
        :param tag: dictionary containing lists of variants and/or test levels
            when only a subset of tests needs to be executed
        """
        super().__init__(
            test_suite_id,
            test_case_id,
            aux_list,
            setup_timeout,
            run_timeout,
            teardown_timeout,
            test_ids,
            tag,
            *args,
            **kwargs,
        )
        if any([setup_timeout, run_timeout, teardown_timeout]):
            log.warning("BasicTestSuiteTeardown does not support test timeouts, it will be discarded")

    def test_suite_tearDown(self):
        """Test method for deconstructing the actual test suite after testing it."""
        pass


class RemoteTestSuiteSetup(BasicTestSuiteSetup):
    """Inherit from unittest testCase and represent setup fixture
    when Message Protocol / TestApp  is used.
    """

    response_timeout = 10

    def __init__(
        self,
        test_suite_id: int,
        test_case_id: int,
        aux_list: Union[List[AuxiliaryInterface], None],
        setup_timeout: Union[int, None],
        run_timeout: Union[int, None],
        teardown_timeout: Union[int, None],
        test_ids: Union[dict, None],
        tag: Union[Dict[str, List[str]], None],
        *args: Any,
        **kwargs: Any,
    ):
        """Initialize Message Protocol / TestApp test-case.

        :param test_suite_id: test suite identification number
        :param test_case_id: test case identification number
        :param aux_list: list of used auxiliaries
        :param setup_timeout: maximum time (in seconds) used to wait
            for a report during setup execution
        :param run_timeout: maximum time (in seconds) used to wait for
            a report during test_run execution
        :param teardown_timeout: the maximum time (in seconds) used to
            wait for a report during teardown execution
        :param test_ids: jama references to get the coverage
            eg: {"Component1": ["Req1", "Req2"], "Component2": ["Req3"]}
        :param tag: dictionary containing lists of variants and/or test levels
            when only a subset of tests needs to be executed
        """
        super().__init__(
            test_suite_id,
            test_case_id,
            aux_list,
            setup_timeout,
            run_timeout,
            teardown_timeout,
            test_ids,
            tag,
            *args,
            **kwargs,
        )
        self.setup_timeout = setup_timeout or RemoteTestSuiteSetup.response_timeout
        self.run_timeout = run_timeout or RemoteTestSuiteSetup.response_timeout
        self.teardown_timeout = teardown_timeout or RemoteTestSuiteSetup.response_timeout

    @test_app_interaction(message_type=message.MessageCommandType.TEST_SUITE_SETUP, timeout_cmd=5)
    def test_suite_setUp(self):
        """Test method for constructing the actual test suite."""
        pass


class RemoteTestSuiteTeardown(BasicTestSuiteTeardown):
    """Inherit from unittest testCase and represent teardown fixture
    when Message Protocol / TestApp is used.
    """

    response_timeout = 10

    def __init__(
        self,
        test_suite_id: int,
        test_case_id: int,
        aux_list: Union[List[AuxiliaryInterface], None],
        setup_timeout: Union[int, None],
        run_timeout: Union[int, None],
        teardown_timeout: Union[int, None],
        test_ids: Union[dict, None],
        tag: Union[Dict[str, List[str]], None],
        *args: Any,
        **kwargs: Any,
    ):
        """Initialize Message Protocol / TestApp test-case.

        :param test_suite_id: test suite identification number
        :param test_case_id: test case identification number
        :param aux_list: list of used auxiliaries
        :param setup_timeout: maximum time (in seconds) used to wait
            for a report during setup execution
        :param run_timeout: maximum time (in seconds) used to wait for
            a report during test_run execution
        :param teardown_timeout: the maximum time (in seconds) used to
            wait for a report during teardown execution
        :param test_ids: jama references to get the coverage
            eg: {"Component1": ["Req1", "Req2"], "Component2": ["Req3"]}
        :param tag: dictionary containing lists of variants and/or test levels
            when only a subset of tests needs to be executed
        """
        super().__init__(
            test_suite_id,
            test_case_id,
            aux_list,
            setup_timeout,
            run_timeout,
            teardown_timeout,
            test_ids,
            tag,
            *args,
            **kwargs,
        )
        self.setup_timeout = setup_timeout or RemoteTestSuiteTeardown.response_timeout
        self.run_timeout = run_timeout or RemoteTestSuiteTeardown.response_timeout
        self.teardown_timeout = teardown_timeout or RemoteTestSuiteTeardown.response_timeout

    @test_app_interaction(
        message_type=message.MessageCommandType.TEST_SUITE_TEARDOWN,
        timeout_cmd=5,
    )
    def test_suite_tearDown(self):
        """Test method for deconstructing the actual test suite after testing it."""
        pass


class BasicTestSuite(unittest.TestSuite):
    """Inherit from the unittest framework test-suite but build it for our integration tests."""

    def __init__(
        self,
        modules_to_add_dir: str,
        test_filter_pattern: str,
        test_suite_id: int,
        *args: Any,
        **kwargs: Any,
    ):
        """Initialize our custom unittest-test-suite.

        .. note::
            1. Will Load from the given path the integration test modules under test
            2. Sort the given test case list by test suite/case id
            3. Place Test suite setup and teardown respectively at top and bottom of test case list
            4. Add sorted test case list to test suite
        """
        # Mother class initialization
        super().__init__(*args, **kwargs)

        # load test from the specified folder
        loader = unittest.TestLoader()
        found_modules = loader.discover(modules_to_add_dir, pattern=test_filter_pattern)

        # sort the test case list by ascendant using test suite and test case id
        test_case_list = sorted(flatten(found_modules), key=tc_sort_key)

        # add sorted test case list to test suite
        self.addTests(test_case_list)

        self.failed_suite_setups = set()

    def check_suite_setup_failed(self, test: BasicTest, result: BannerTestResult) -> None:
        """Check if the suite setup has failed and store failed suite id.
        Search in the global unittest result object, which save all the results
        of the tests performed up to that point, if the suite setup that ran
        has failed then store the suite id.

        :param test: test to check
        :param result: unittest result object
        """
        if isinstance(test, BasicTestSuiteSetup):
            if result.error_occurred:
                self.failed_suite_setups.add(test.test_suite_id)

    def run(self, result: BannerTestResult, debug: bool = False) -> BannerTestResult:
        """Override run method from unittest.suite.TestSuite.
        Added functionality:
        Skip suite tests if the parent test suite setup has failed.

        :param result: unittest result storage
        :param debug: True to enter debug mode, defaults to False
        :return: test suite result
        """
        topLevel = False
        if getattr(result, "_testRunEntered", False) is False:
            result._testRunEntered = topLevel = True

        for index, test in enumerate(self):
            if result.shouldStop:  # pragma: no cover
                break

            if _isnotsuite(test):
                self._tearDownPreviousClass(test, result)
                self._handleModuleFixture(test, result)
                self._handleClassSetUp(test, result)
                result._previousTestClass = test.__class__

                if getattr(test.__class__, "_classSetupFailed", False) or getattr(
                    result, "_moduleSetUpFailed", False
                ):  # pragma: no cover
                    continue

            if not debug:
                if test.test_suite_id in self.failed_suite_setups:
                    result.addSkip(
                        test,
                        f"Suite Setup failed for test suite {test.test_suite_id}",
                    )
                else:
                    test(result)
                    self.check_suite_setup_failed(test, result)

            else:  # pragma: no cover
                test.debug()

            if self._cleanup:
                self._removeTestAtIndex(index)

        if topLevel:
            self._tearDownPreviousClass(None, result)
            self._handleModuleTearDown(result)
            result._testRunEntered = False
        return result


def tc_sort_key(tc):
    """Sort-key for testcases.

    will sort by test-suite/test-case, but the setup will always be first,
    the teardown will always be last.

    :param tc: a Base or Remote TestSuite/TestCase to rank

    :return: key for :py:func:`sorted`

    :raises: any exception that occurs during test loading
    """
    fix_ind = 0
    if isinstance(tc, (BasicTestSuiteSetup, RemoteTestSuiteSetup)):
        fix_ind = -1
    elif isinstance(tc, (BasicTestSuiteTeardown, RemoteTestSuiteTeardown)):
        fix_ind = 1
    elif isinstance(tc, unittest.loader._FailedTest):
        raise tc._exception
    return (fix_ind, tc.test_suite_id, tc.test_case_id)


def flatten(it: unittest.TestSuite) -> Iterable[BasicTest]:
    """Flatten all level of nesting.

    :param it: nested iterable
    :return: first not nested items
    """
    for x in it:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Can Communication Channel using PCAN hardware
*********************************************

:module: cc_pcan_can

:synopsis: CChannel implementation for CAN(fd) using PCAN API from python-can

.. currentmodule:: cc_pcan_can

"""
import logging
import os
import shutil
import sys
from pathlib import Path
from typing import Dict, List, Literal, Optional, Union

try:
    import can
    import can.bus
    import can.interfaces.pcan.basic as PCANBasic
    from can import TRCFileVersion
except ImportError as e:
    raise ImportError(f"{e.name} dependency missing, consider installing pykiso with 'pip install pykiso[can]'")

try:
    # use to avoid timestamp inconsistances between pykiso and python can
    import uptime

    if uptime.boottime() is None:
        boottime_epoch = 0
    else:
        boottime_epoch = uptime.boottime().timestamp()
except ImportError:
    boottime_epoch = 0

from pykiso import CChannel

from .trc_handler import TRCReaderCanFD, TRCWriterCanFD, TypedMessage

log = logging.getLogger(__name__)


class PcanFilter(logging.Filter):
    """Filter specific pcan logging messages"""

    def filter(self, record: logging.LogRecord) -> bool:
        """Determine if the specified record is to be logged. It will not if
        it is a pcan bus error message

        :param record: record of the event to filter if it is a pcan bus error

        :return: True if the record should be logged, or False otherwise.
        """
        return not record.getMessage().startswith("Bus error: an error counter")


class CCPCanCan(CChannel):
    """CAN FD channel-adapter."""

    def __init__(
        self,
        interface: str = "pcan",
        channel: str = "PCAN_USBBUS1",
        state: str = "ACTIVE",
        trace_path: str = "",
        trace_size: int = 100,  # 100 is maximum size
        bitrate: int = 500000,
        is_fd: bool = True,
        enable_brs: bool = False,
        f_clock_mhz: int = 80,
        nom_brp: int = 2,
        nom_tseg1: int = 63,
        nom_tseg2: int = 16,
        nom_sjw: int = 16,
        data_brp: int = 4,
        data_tseg1: int = 7,
        data_tseg2: int = 2,
        data_sjw: int = 2,
        is_extended_id: bool = False,
        remote_id: int = None,
        can_filters: list = None,
        logging_activated: bool = True,
        bus_error_warning_filter: bool = False,
        merge_trc_logs: bool = True,
        strategy_trc_file: Optional[Literal["testRun", "testCase"]] = None,
        **kwargs,
    ):
        """Initialize can channel settings.

        :param interface: python-can interface modules used
        :param channel: the can interface name
        :param state: BusState of the channel
        :param trace_path: path to write the trace (can be a folder or a .trc file)
            .. note:: If the .trc file already exists, it will be overwritten. If the trace_path
            is an existing folder, a default name will be generated for the trace file
            containing the timestamp. If the trace_path is a non-existent folder, the
            folder will be created and the default name will be used for the trace
            file. If the trace_path is not defined by the user, the default file will be
            saved in the current working directory.
        :param trace_size: maximum size of the trace (in MB)
        :param bitrate: Bitrate of channel in bit/s,ignored if using CanFD
        :param is_fd: Should the Bus be initialized in CAN-FD mode
        :param enable_brs: sets the bitrate_switch flag to use higher transmission speed
        :param f_clock_mhz:  Clock rate in MHz
        :param nom_brp: Clock prescaler for nominal time quantum
        :param nom_tseg1: Time segment 1 for nominal bit rate, that is,
            the number of quanta from the Sync Segment to the sampling point
        :param nom_tseg2: Time segment 2 for nominal bit rate,
            that is, the number of quanta from the sampling point to the end of the bit
        :param nom_sjw: Synchronization Jump Width for nominal bit rate.
            Decides the maximum number of time quanta that the controller
            can resynchronize every bit
        :param data_brp: Clock prescaler for fast data time quantum
        :param data_tseg1: Time segment 1 for fast data bit rate, that is,
            the number of quanta from the Sync Segment to the sampling point
        :param data_tseg2: Time segment 2 for fast data bit rate, that is,
            the number of quanta from the sampling point to the end of the bit.
            In the range (1..16)
        :param data_sjw: Synchronization Jump Width for fast data bit rate
        :param is_extended_id: This flag controls the size of the arbitration_id field
        :param remote_id: id used for transmission
        :param can_filters: iterable used to filter can id on reception
        :param logging_activated: boolean used to disable logfile creation
        :param bus_error_warning_filter: if True filter the PCAN driver warnings
            'Bus error: an error counter' from the logs.
        :param merge_trc_logs: if True, merge all traces in one file at the end of the program
        :param strategy_trc_file: Strategy for the trace file by default (set to None) it will
            be one trace file for all the tests run, if set to 'test' it will be one trace file
            per test run and 'testCase' it will be one trace file per testCase.
        """
        super().__init__(**kwargs)
        self.interface = interface
        self.channel = channel
        self.state = can.bus.BusState[state.upper()]
        self.trace_path = Path(trace_path)
        self.trace_size = trace_size
        self.bitrate = bitrate
        self.is_fd = is_fd
        self.enable_brs = enable_brs
        self.f_clock_mhz = f_clock_mhz
        self.nom_brp = nom_brp
        self.nom_tseg1 = nom_tseg1
        self.nom_tseg2 = nom_tseg2
        self.nom_sjw = nom_sjw
        self.data_brp = data_brp
        self.data_tseg1 = data_tseg1
        self.data_tseg2 = data_tseg2
        self.data_sjw = data_sjw
        self.is_extended_id = is_extended_id
        self.remote_id = remote_id
        self.can_filters = can_filters
        self.bus = None
        self.logging_activated = logging_activated
        self.raw_pcan_interface = None
        self.strategy_trc_file = strategy_trc_file
        # Set a timeout to send the signal to the GIL to change thread.
        # In case of a multi-threading system, all tasks will be called one after the other.
        self.timeout = 1e-6
        self.trc_count = 0
        self.boottime_epoch = boottime_epoch
        self._initialize_trace()
        self.merge_trc_logs = merge_trc_logs if strategy_trc_file is None else False
        self._trc_file_names: dict[Path, str | None] = {}
        self.trace_running = False
        self.opened = False
        if bus_error_warning_filter:
            logging.getLogger("can.pcan").addFilter(PcanFilter())

        if self.enable_brs and not self.is_fd:
            log.internal_warning("Bitrate switch will have no effect because option is_fd is set to false.")

    def _initialize_trace(self) -> None:
        """Initialize the trace path and check its size

        :raises ValueError: if the trace_path is a file but not a trc file
        """
        # Handle trace path and name
        if self.trace_path.suffix == ".trc":
            self.trace_name = self.trace_path.name
            self.trace_path = self.trace_path.parent
        elif self.trace_path and self.trace_path.suffix == "":
            self.trace_name = None
        elif not self.trace_path:
            self.trace_path = Path.cwd()
            self.trace_name = None
        elif self.trace_path.suffix not in [".trc", ""]:
            raise ValueError(f"Trace name {self.trace_path.name} is incorrect, it should be a trc file")

        # Check trace size
        if not (0 < self.trace_size <= 100):
            self.trace_size = 100
            log.internal_warning(
                "Make sure trace size is between 1 and 100 Mb. Setting trace size to default value "
                f"value : {self.trace_size}."
            )

    def _cc_open(self) -> None:
        """Open a can bus channel, set filters for reception and activate PCAN log."""
        if self.opened:
            log.warning("Pcan is already opened")
            return
        self.bus = can.interface.Bus(
            interface=self.interface,
            channel=self.channel,
            state=self.state,
            bitrate=self.bitrate,
            fd=self.is_fd,
            f_clock_mhz=self.f_clock_mhz,
            nom_brp=self.nom_brp,
            nom_tseg1=self.nom_tseg1,
            nom_tseg2=self.nom_tseg2,
            nom_sjw=self.nom_sjw,
            data_brp=self.data_brp,
            data_tseg1=self.data_tseg1,
            data_tseg2=self.data_tseg2,
            data_sjw=self.data_sjw,
            can_filters=self.can_filters,
        )

        if self.logging_activated and self.raw_pcan_interface is None:
            self.raw_pcan_interface = PCANBasic.PCANBasic()
            if self.strategy_trc_file is None:
                self._pcan_configure_trace()
        self.opened = True

    def _pcan_configure_trace(self) -> None:
        """Configure PCAN dongle to create a trace file.

        If self.trace_path is set, this path will be created, if it does not
        exist and the logfile will be placed there.
        Otherwise it will be logged to the current working directory if a
        default filename, which will be overwritten in successive calls.
        If an error occurs, the trace will not be started and the error logged.
        No exception is thrown in this case.
        """
        pcan_channel = getattr(PCANBasic, self.channel)
        if self.trace_path is None:
            log.internal_warning("No trace path specified, an existing trace will be overwritten.")
            pcan_trace_configuration = PCANBasic.TRACE_FILE_OVERWRITE
        else:
            pcan_trace_configuration = PCANBasic.TRACE_FILE_DATE | PCANBasic.TRACE_FILE_TIME

        try:
            if self.trace_path is not None:
                if not Path(self.trace_path).exists():
                    Path(self.trace_path).mkdir(parents=True, exist_ok=True)
                    log.internal_info("Path %s created", self.trace_path)
                self._pcan_set_value(
                    pcan_channel,
                    PCANBasic.PCAN_TRACE_LOCATION,
                    bytes(self.trace_path),
                )
                log.internal_info(
                    "Tracefile path in PCAN device configured to %s",
                    self.trace_path,
                )

            if sys.platform != "darwin":
                log.internal_info("Segmented option of trace file activated.")
                pcan_trace_configuration |= PCANBasic.TRACE_FILE_SEGMENTED

                if self.trace_size != 10:
                    log.internal_info("Trace size set to %d MB.", self.trace_size)
                    self._pcan_set_value(
                        pcan_channel,
                        PCANBasic.PCAN_TRACE_SIZE,
                        self.trace_size,
                    )
            else:
                log.internal_warning("TRACE_FILE_SEGMENTED deactivated for macos!")
                pcan_trace_configuration |= PCANBasic.TRACE_FILE_SINGLE

            self._pcan_set_value(
                pcan_channel,
                PCANBasic.PCAN_TRACE_CONFIGURE,
                pcan_trace_configuration,
            )
            log.internal_info("Tracefile configured")

            self._pcan_set_value(
                pcan_channel,
                PCANBasic.PCAN_TRACE_STATUS,
                PCANBasic.PCAN_PARAMETER_ON,
            )
            self.trace_running = True
            log.internal_info("Trace activated")
            # Add the file name and trace path to rename them later
            if self._trc_file_names.get(self.trace_path, None) is None:
                self._trc_file_names[self.trace_path] = []
            self._trc_file_names[self.trace_path].append(self.trace_name)
            self.trc_count += 1
        except RuntimeError:
            log.error(f"Logging for {self.channel} not activated")
        except OSError as e:
            log.error(f"Can not create log folder for PCAN logs: {e}")
            log.error(f"Logging for {self.channel} not activated")

    def _pcan_set_value(self, channel, parameter, buffer) -> None:
        """Set a value in the PCAN api.

        If this is not successful, a RuntimeError is returned, as well as the
        PCAN error text is logged, if possible.

        :param channel: Channel for PCANBasic.SetValue
        :param parameter: Parameter for PCANBasic.SetValue
        :param buffer: Buffer for PCANBasic.SetValue

        :raises RuntimeError: Raised if the function is not successful
        """
        try:
            result = self.raw_pcan_interface.SetValue(
                channel,
                parameter,
                buffer,
            )
        except Exception as e:
            log.error(f"Exception in call to SetValue: {e}")
            raise RuntimeError("Error configuring logging on PCAN")
        else:
            if result != PCANBasic.PCAN_ERROR_OK:
                _, error_msg = self.raw_pcan_interface.GetErrorText(result)
                log.error(error_msg)
                raise RuntimeError(f"Error configuring logging on PCAN: {result}")

    def _cc_close(self) -> None:
        """Close the current can bus channel and uninitialize PCAN handle."""
        if not self.opened:
            log.warning("Pcan is already closed")
            return
        self.stop_pcan_trace()
        self.bus.shutdown()
        self.bus = None
        if self.logging_activated:
            try:
                result = self.raw_pcan_interface.Uninitialize(PCANBasic.PCAN_NONEBUS)
            except Exception as e:
                log.exception(f"Error in call to Uninitialize: {e}")
            else:
                if result != PCANBasic.PCAN_ERROR_OK:
                    _, error_msg = self.raw_pcan_interface.GetErrorText(result)
                    log.error(error_msg)
            finally:
                self.raw_pcan_interface = None
        self.opened = False

    def _cc_send(self, msg: bytes, remote_id: Optional[int] = None, **kwargs) -> None:
        """Send a CAN message at the configured id.

        If remote_id parameter is not given take configured ones

        :param msg: data to send
        :param remote_id: destination can id used
        :param kwargs: named arguments
        """

        remote_id = remote_id or self.remote_id

        can_msg = can.Message(
            arbitration_id=remote_id,
            data=msg,
            is_extended_id=self.is_extended_id,
            is_fd=self.is_fd,
            bitrate_switch=self.enable_brs,
        )
        self.bus.send(can_msg)

        log.internal_debug("%s sent CAN Message: %s, data: %s", self, can_msg, msg)

    def _cc_receive(self, timeout: float = 0.0001) -> Dict[str, Union[bytes, int, None]]:
        """Receive a can message using configured filters.

        :param timeout: timeout applied on reception

        :return: the received data and the source can id
        """
        try:  # Catch bus errors & rcv.data errors when no messages where received
            received_msg = self.bus.recv(timeout=timeout or self.timeout)

            if received_msg is not None:
                frame_id = received_msg.arbitration_id
                payload = received_msg.data
                timestamp = received_msg.timestamp - self.boottime_epoch

                log.internal_debug(
                    "received CAN Message: %s, %s, %s",
                    frame_id,
                    payload,
                    timestamp,
                )
                return {
                    "msg": payload,
                    "remote_id": frame_id,
                    "timestamp": timestamp,
                }
            else:
                return {"msg": None}
        except can.CanError as can_error:
            log.internal_info("encountered CAN error while receiving message: %s", can_error)
            return {"msg": None}
        except Exception:
            log.exception(f"encountered error while receiving message via {self}")
            return {"msg": None}

    @staticmethod
    def _extract_header(trace: Path) -> str:
        """Extract data from trc file header

        :param trace: trc file from which to extract the data

        :return: header data
        """
        header_data = ""

        with trace.open("r") as trc:
            data = trc.read().splitlines(True)
            for line in data:
                if line.startswith(";"):
                    header_data += line
                else:
                    break
        return header_data

    def _merge_trc(self) -> None:
        """Merge all traces file in one and fix potential inconsistencies."""

        list_of_traces = []
        first_trace_name = None
        first_trace_path = None

        for trace_path, trace_file_names in self._trc_file_names.items():
            if isinstance(trace_path, str):
                trace_path = Path(trace_path)
            if first_trace_path is None:
                first_trace_name = trace_file_names[0]
                first_trace_path = trace_path
            # Get all the trace files created in a list
            list_of_traces.extend(sorted(trace_path.glob("*.trc"), key=os.path.getmtime)[-len(trace_file_names) :])

        try:
            if first_trace_name is None:
                # If a log file is not provided, merge everything into the first created trace
                result_trace = list_of_traces[0]
            else:
                # Otherwise create a separate file
                result_trace = Path(first_trace_path / first_trace_name)
                # replace the first trace with the result trace in the trace list
                # to ensure that all traces except the merged trace are deleted in _read_trace_messages
                list_of_traces[0] = Path(shutil.move(str(list_of_traces[0]), str(result_trace)))

            # Extract header
            header_data = CCPCanCan._extract_header(list_of_traces[0])

            # Get messages from all traces
            try:
                messages = self._read_trace_messages(list_of_traces, result_trace)
            except ValueError:
                return

            # Write header in result trace
            result_trace.write_text(header_data)

            # Write all messages in result trace
            with TRCWriterCanFD(result_trace) as writer:
                writer.file_version = self.trc_file_version
                log.internal_debug("TRC file version is %s", self.trc_file_version)
                writer.header_data = header_data
                for msg in messages:
                    writer.on_message_received(msg, self.trc_start_time.timestamp())

        except IndexError:
            log.internal_warning("No trace to merge")

    def _read_trace_messages(self, list_of_traces: List[Path], result_trace: Path) -> List[TypedMessage]:
        """Get the list of messages for all traces with corrected offset

        :param list_of_traces: list of all traces
        :param result_trace: trace where to merge

        :return: list of corrected typed messages from all traces
        """
        messages = []
        for trc in list_of_traces:
            log.internal_debug("Merging trace %s into %s", trc.name, result_trace.name)
            with TRCReaderCanFD(trc) as reader:
                current_trc_messages = list(reader)

                if trc == result_trace:
                    # Get info for the first trace
                    self.trc_start_time = reader.start_time
                    self.trc_file_version = reader.file_version
                else:
                    # Get offset in milliseconds
                    current_trc_start_time = reader.start_time
                    offset = (current_trc_start_time - self.trc_start_time).total_seconds()
                    current_trc_messages = CCPCanCan._remove_offset(current_trc_messages, offset)
                    os.remove(trc)
                if self.trc_file_version in [
                    TRCFileVersion.V1_1,
                    TRCFileVersion.V1_2,
                    TRCFileVersion.V1_3,
                ]:
                    log.warning(
                        "Trace merging is not available for trc file version %s",
                        self.trc_file_version,
                    )
                    raise ValueError
                messages += current_trc_messages
        return messages

    @staticmethod
    def _remove_offset(messages: List[TypedMessage], offset: float) -> List[TypedMessage]:
        """Remove offset to the timestamp of a list of messages

        :param messages: list of messages to adapt
        :param offset: offset to add in seconds

        :return: List of message with corrected timestamps
        """
        for msg in messages:
            msg.timestamp = msg.timestamp + offset
        return messages

    def _rename_trc(self):
        """Rename the trace file created if a name has been specified"""
        for trace_path, trace_file_names in self._trc_file_names.items():
            list_of_traces = sorted(trace_path.glob("*.trc"), key=os.path.getmtime)[-len(trace_file_names) :]
            for index, file_name in enumerate(trace_file_names):
                if file_name is not None:
                    list_of_traces[index].rename(trace_path / file_name)

    def shutdown(self) -> None:
        """Destructor method."""
        if not self.logging_activated:
            return
        if self.merge_trc_logs:
            self._merge_trc()
        else:
            self._rename_trc()

    def stop_pcan_trace(self):
        """
        Stops the PCAN trace if it is currently running.
        :return: None
        """
        if not self.logging_activated:
            return
        if not self.trace_running:
            log.warning("Trace is already stopped")
            return

        pcan_channel = getattr(PCANBasic, self.channel)
        self._pcan_set_value(
            pcan_channel,
            PCANBasic.PCAN_TRACE_STATUS,
            PCANBasic.PCAN_PARAMETER_OFF,
        )
        self.trace_running = False

    def start_pcan_trace(self, trace_path: Optional[str] = None, trace_size: int = None) -> None:
        """Start the PCAN trace, the trace file will be renamed after the pcan trace will be stopped.

        :param trace_path: Trace path where the trace should be written if None is given it will use
            the trace path defined for the last trace created, defaults to None
        :param trace_size:  maximum size of the trace (in MB), defaults to None
        """
        if not self.logging_activated:
            log.warning("Logging is not activated")
            return
        if self.trace_running:
            log.warning("Trace is already started")
            return

        self.trace_size = trace_size or self.trace_size
        self.trace_path = Path(trace_path) if trace_path else self.trace_path
        self._initialize_trace()
        self._pcan_configure_trace()



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Is used when pykiso is called as a module.

It redirects to pykiso.cli.main.

.. command-output:: python -m pykiso --help
    :ellipsis: 4



:module: __main__

"""

from .cli import main

if __name__ == "__main__":
    main()



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Text Test Result with banners
*****************************

:module: text_result

:synopsis: implements a test result that displays the test execution
    wrapped in banners.

.. currentmodule:: text_result
"""

from __future__ import annotations

import logging
import os
import sys
import textwrap
import time
import typing
from contextlib import nullcontext
from pathlib import Path
from shutil import get_terminal_size
from typing import List, Optional, TextIO, Union
from unittest import TextTestResult
from unittest.case import _SubTest

from ..test_coordinator.test_case import BasicTest
from ..test_coordinator.test_suite import BaseTestSuite

if typing.TYPE_CHECKING:
    from pykiso.types import ExcInfoType, PathType

log = logging.getLogger(__name__)


class MultiFileHandler:
    """Class to write simultaneously to multiple files."""

    def __init__(self, files: List[Path] | None = None):
        self.files: dict[Path, TextIO] = {}
        if files:
            for file in files:
                self.add_file(file)

    def add_file(self, file_path: Path):
        """Add a file by path to the handler."""
        if file_path not in self.files:
            f = open(file_path, "a", encoding="utf-8")
            self.files[file_path.resolve()] = f

    def remove_file(self, file_path: Path):
        """Remove a file by path from the handler."""
        f = self.files.pop(file_path.resolve(), None)
        if f and hasattr(f, "close"):
            f.close()

    def write(self, message: str):
        for f in self.files.values():
            f.write(message)

    def flush(self):
        for f in self.files.values():
            f.flush()
            os.fsync(f.fileno())

    def close(self):
        for f in self.files.values():
            if hasattr(f, "close"):
                f.close()
        self.files.clear()


class ResultStream:
    """Class that duplicates sys.stderr to a log file if a file path is provided.

    When passed to a TestRunner or a TestResult, this allows to display the
    information from the test run in the log file.
    """

    def __new__(cls, file: Optional[PathType] = None):
        """Customize class creation to return an instance of this class
        if a file path is provided, or simply ``sys.stderr`` if no file
        path is provided.

        :param file: the file to write stderr to.
        :return: an instance of this class or ``sys.stderr``.
        """
        # don't bother instantiating and simply return stderr as context manager
        if file is None:
            return nullcontext(sys.stderr)
        # otherwise return a real instance
        return super().__new__(cls)

    def __init__(self, file: Optional[PathType]):
        """Initialize the streams.

        :param file: file where stderr should be written.
        """
        self.stderr = sys.stderr
        self.multifile_handler = MultiFileHandler([Path(file) if isinstance(file, str) else file])
        sys.stderr = self

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def write(self, message: str):
        self.stderr.write(message)
        self.multifile_handler.write(message)

    def flush(self):
        self.stderr.flush()
        self.multifile_handler.flush()

    def close(self):
        """Close or restore each stream."""
        if self.stderr is not None:
            sys.stderr = self.stderr
            self.stderr = None

        if self.multifile_handler is not None:
            self.multifile_handler.close()
            self.multifile_handler = None


class BannerTestResult(TextTestResult):
    """TextTestResult subclass showing results wrapped in banners (frames)."""

    BANNER_CHAR_WIDTH = 4

    def __init__(self, stream: TextIO, descriptions: bool, verbosity: int):
        """Constructor. Initialize TextTestResult and the banner's width.

        The banner's width is set to the terminal size. In the case
        where this fails the fallback width corresponds to the default
        width of a Jenkins "console".

        :param stream: stream to print the result information
            (default: stderr)
        :param descriptions: unused (required for TextTestResult)
        :param verbosity: unused (required for TextTestResult)
        """
        super().__init__(stream, descriptions, verbosity)
        # to determine whether the test succeeded or failed
        self._error_occurred = False
        # fallback is the default width in Jenkins
        size = get_terminal_size(fallback=(150, 24))
        # avoid border effects due to newlines
        self.width = size.columns - 1
        self.successes: List[Union[BasicTest, BaseTestSuite]] = []

    @property
    def error_occurred(self):
        return self._error_occurred

    def _banner(
        self,
        text: Union[List, str],
        width: Optional[int] = None,
        sym: str = "#",
    ) -> str:
        """Format the provided text within a frame composed of the provided
        symbol.

        Works with multiline strings (either as a string containing
        newlines or split into a list with one entry per line).

        :param text: text to format
        :paran width: width of the banner
        :param sym: symbol used to compose the banner
        :return: the text enclosed in a banner
        """
        width = width or self.width
        line = sym * width
        if isinstance(text, str):
            text = text.split("\n")
        if isinstance(text, list):
            text = "\n".join(f"{sym} {line: <{width-self.BANNER_CHAR_WIDTH}} {sym}" for line in text)  # noqa: E226
        banner = f"{line}\n{text}\n{line}\n"
        return banner

    def getDescription(self, test: Union[BasicTest, BaseTestSuite]) -> str:
        """Return the entire test method docstring.

        :param test: running testcase
        :return: the wrapped docstring
        """
        doc = ""
        if getattr(test, "_testMethodDoc", None) is not None:
            for line in test._testMethodDoc.splitlines():
                doc += "\n" + textwrap.fill(line.strip(), width=self.width - self.BANNER_CHAR_WIDTH)
        return doc

    def print_skipped(self, test: Union[BasicTest, BaseTestSuite]) -> None:
        """Print a banner for a skipped test with the reason why
        it was skipped.

        :param test: the running testcase
        """
        # gather test module, test class, test method and test description
        module_name = test.__module__
        test_name = str(test)
        top_str = "SKIPPED TEST: "
        is_skipped = True
        if len(module_name + test_name) < self.width - len(top_str):
            test_name = f"{module_name}.{test_name}"
        else:
            test_name += f"\nmodule: {module_name}"
        reason = f"Reason: {test.__unittest_skip_why__}"
        # create and print test start banner
        top_str += f"{test_name}\n{reason}"
        top_banner = self._banner(top_str)
        if is_skipped:
            top_banner += "\n"
        self.stream.write(top_banner)
        self.stream.flush()

    def startTest(self, test: Union[BasicTest, BaseTestSuite]) -> None:
        """Print a banner containing the test information and the test
        method docstring when starting a test case.

        :param test: testcase that is about to be run
        """
        super().startTest(test)
        self._error_occurred = False
        # gather test module, test class, test method and test description
        module_name = test.__module__
        test_name = str(test)
        addendum = ""
        doc = self.getDescription(test).rstrip()
        top_str = "RUNNING TEST: "
        if len(module_name + test_name) < self.width - len(top_str):
            test_name = f"{module_name}.{test_name}"
        else:
            addendum += f"\nmodule: {module_name}"
        # create start banner
        top_str += f"{test_name}{addendum}{doc}"
        top_banner = self._banner(top_str)
        # print it
        self.stream.write(top_banner)
        self.stream.flush()
        # start monitoring test duration
        test.start_time = time.time()

    def stopTest(self, test: Union[BasicTest, BaseTestSuite]) -> None:
        """Print a banner containing the test information and its result.

        :param test: terminated testcase
        """
        # print a banner indicating that the test was skipped
        # this needs to be done here as startTest is not called for a skipped test with python>=3.12
        if getattr(test, "__unittest_skip__", False):
            self.print_skipped(test)
            return super().stopTest(test)

        test.stop_time = time.time()
        test.elapsed_time = test.stop_time - test.start_time
        result = "FAILED" if self.error_occurred else "PASSED"
        bot_str = f"END OF TEST: {test}"
        result_str = f"  ->  {result} in {test.elapsed_time:.3f}s"  # noqa: E221,E222,E231

        if len(bot_str + result_str) < self.width - self.BANNER_CHAR_WIDTH:
            bot_str += result_str
        else:
            bot_str += "\n" + result_str

        bot_banner = self._banner(bot_str) + "\n"
        self.stream.write(bot_banner)
        self.stream.flush()
        super().stopTest(test)

    def addFailure(self, test: Union[BasicTest, BaseTestSuite], err: ExcInfoType) -> None:
        """Set the error flag when a failure occurs in order to get the
        individual test case result.

        :param test: testcase which failure will be reported
        :param err: tuple returned by sys.exc_info
        """
        super().addFailure(test, err)
        self._error_occurred = True

    def addSuccess(self, test: Union[BasicTest, BaseTestSuite]) -> None:
        """Add a testcase to the list of succeeded test cases.

        :param test: running testcase that succeeded
        """
        self.successes.append(test)

    def addError(self, test: Union[BasicTest, BaseTestSuite], err: ExcInfoType) -> None:
        """Set the error flag when an error occurs in order to get the
        individual test case result.

        :param test: running testcase that errored out
        :param err: tuple returned by sys.exc_info
        """
        super().addError(test, err)
        self._error_occurred = True

    def addSubTest(
        self,
        test: Union[BasicTest, BaseTestSuite],
        subtest: _SubTest,
        err: ExcInfoType,
    ) -> None:
        """Set the error flag when an error occurs in a subtest.

        :param test: running testcase
        :param subtest: subtest ran
        :param err: tuple returned by sys.exc_info
        """
        super().addSubTest(test, subtest, err)
        if err is not None:
            self._error_occurred = True

    def printErrorList(self, flavour: str, errors: List[tuple]):
        """Print all errors at the end of the whole tests execution.

        Overwrites the unittest method to have a nicer output.

        :param flavour: failure reason
        :param errors: list of failed tests with their error message
        """

        for test, err in errors:
            self.stream.writeln(self.separator1)
            self.stream.writeln("%s" % test)
            self.stream.writeln("%s" % self.getDescription(test))
            self.stream.writeln("%s: %s" % (flavour, err))



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

import pathlib
import re
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import click
import isort
from black import FileMode, format_str
from jinja2 import Environment, FileSystemLoader

from pykiso import config_parser

ROOT_PATH = pathlib.Path(__file__).parent.resolve()


def format_value(value: Any) -> Union[Any, str]:
    """Add double quotes to a string if value is
    of type string. Else return original value

    :param value: value to evaluate
    :return: quoted string if value was a string
    """

    if isinstance(value, (str)):
        return f'"{value}"'

    return value


def find_string_in(nested_dict: dict, search_value: str, prepath: Tuple[str, ...] = ()) -> Optional[Tuple[str, ...]]:
    """Find if string is part of a value in a nested dictionary.

    :param nested_dict: dictionary to search in
    :param search_value: string to search for
    :param prepath: path storage for recursive call, defaults to ()
    :return: path to locataion when found else None
    """
    for key, value in nested_dict.items():
        path = prepath + (key,)
        if isinstance(value, str) and search_value in value:  # found value
            return path
        elif isinstance(value, dict):
            key_path = find_string_in(value, search_value, path)
            if key_path is not None:
                return key_path
    # If not found
    return None


def nested_get(dic: dict, keys: List[str], copy: bool = False) -> Any:
    """Get dictionary items from a sequence of keys

    :param dic: dictionary to search in
    :param keys: key path as list
    :param copy: if true return a new dictionary instance
    :return: dictionary value at located key positions
    """
    for key in keys:
        dic = dic[key]

    if copy:
        return dict(dic)

    return dic


def remove_key(dic: Dict[str, Union[Dict, Any]], key_path: List[str]) -> None:
    """Remove dictionary element at given location.

    :param dic: dictionary where the key shall be removed
    :param key_path: key path as list
    """
    for key in key_path[:-1]:
        dic = dic[key]
    del dic[key_path[-1]]


def get_imports(file: str) -> List[str]:
    """Get all "from x import y" from a python script.

    :param file: python file to read in
    :return: all from imports
    """

    return [line for line in open(file, encoding="utf8") if re.match(r"from.*$", line)]


@click.command()
@click.argument("filename", type=click.Path(exists=True, dir_okay=False))
@click.option(
    "-d",
    "--destination",
    default="./conftest.py",
    help="Output file path.",
    type=click.Path(writable=True),
    show_default=True,
)
def main(filename: str, destination: str):
    """Generate files from a pykiso config to use
    pytest instead of pykiso tests.

    :param filename: pykiso yaml file to convert
    :param destination: destination file path
    """
    pykiso_config = config_parser.parse_config(filename)

    template_loader = FileSystemLoader(searchpath=str(ROOT_PATH / "templates"))
    # The issue B701 wants autoescape to be set to true in order to filter input strings to escape any HTML content submitted via template variables.
    # Autoescape needs to be switched off, since executable python code is included which falsely triggers the issue.
    template_env = Environment(loader=template_loader, autoescape=False)  # nosec B701
    template_env.filters["format_value"] = format_value
    template = template_env.get_template("conftest_template.jinja2")

    # Grab import information -> Channels and Auxiliaries
    yaml_config_folder = pathlib.Path(filename).resolve().parent

    import_info = set()

    for _, value in pykiso_config["auxiliaries"].items():
        import_info.add(value["type"])

    for _, value in pykiso_config["connectors"].items():
        import_info.add(value["type"])

    module_imports = {}
    for aux_con_type in import_info:
        value, key = aux_con_type.split(":")[-2:]
        if ".py" in value:
            ext_file_imports = get_imports(yaml_config_folder / value)
            for imp in ext_file_imports:
                _, value, _, key = imp.strip().split(" ")

        module_imports[key] = value

    # Check if we have a proxy auxiliary
    key_path = find_string_in(pykiso_config, "ProxyAuxiliary")
    proxy_aux_config = None
    if key_path:
        proxy_aux_path = list(key_path[:-1])
        proxy_aux_name = key_path[1]
        proxy_aux_config = nested_get(pykiso_config, proxy_aux_path, copy=True)
        remove_key(pykiso_config, proxy_aux_path)
        proxy_aux_config["name"] = proxy_aux_name

    # Use jinja to render conftest
    conftest = template.render(
        sys_args=sys.argv,
        dict_item=pykiso_config,
        proxy_aux_config=proxy_aux_config,
        module_imports=module_imports,
    )
    # Sort imports with isort and use black to format the code
    conftest = format_str(isort.code(conftest), mode=FileMode())

    destination_path = Path(destination)
    if destination_path.is_dir():
        destination_path = destination_path / "conftest.py"
    else:
        if not destination_path.suffix == ".py":
            destination_path = destination_path.with_suffix(".py")

    with open(destination_path, "w", encoding="utf8") as file:
        file.write(conftest)

    print(f"Conftest has been written to {destination_path.resolve()}")


if __name__ == "__main__":
    main()



##########################################################################
# Copyright (c) 2010-2023 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
pykiso plugin for pytest
************************

:module: pytest

:synopsis: integrate pykiso features within pytest.

.. currentmodule:: pytest

"""

from . import collection, commandline, hooks, logging, markers, reporting
from .collection import *
from .commandline import *
from .logging import *
from .markers import *
from .reporting import *



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Handle common communication with device under test
**************************************************

When using a Remote TestCase/TestSuite, the integration test framework
handles internal messaging and control flow using a message format
defined in :py:class:`pykiso.Message`.

:py:mod:`pykiso.test_message_handler` defines the messaging protocol
from a behavioral point of view.

:module: test_message_handler

:synopsis: default communication between TestManagement and DUT.

.. currentmodule:: test_message_handler

"""

import functools
import logging
from typing import Callable

from pykiso import message

log = logging.getLogger(__name__)


def test_app_interaction(message_type: message.MessageCommandType, timeout_cmd: int = 5) -> Callable:
    """Handle test app basic interaction depending on the decorated
    method.

    :param message_type: message command sub-type (test case/suite
        run, setup, teardown....)
    :param timeout_cmd: timeout in seconds for auxiliary run_command

    :return: inner decorator function
    """

    def inner_interaction(func: Callable) -> Callable:
        """Inner decorator function.

        :param func: decorated method

        :return: decorator inner function
        """

        @functools.wraps(func)
        def handle_interaction(self, *args: tuple, **kwargs: dict) -> None:
            """Handle Test App communication mechanism during for all
            available fixtures(setup, run, teardown,...)

            :param args: positional arguments
            :param kwargs: named arguments

            :return: decorated method return (actual case None)
            """

            fixture = func.__name__.lower()

            if "setup" in fixture and "suite" in fixture:
                log.info(f"--------------- SUITE SETUP: {self.test_suite_id} ---------------")
                timeout_resp = self.setup_timeout
            elif "setup" in fixture:
                log.info(f"--------------- SETUP: {self.test_suite_id}, {self.test_case_id} ---------------")
                timeout_resp = self.setup_timeout
            elif "run" in fixture:
                log.info(f"--------------- RUN: {self.test_suite_id}, {self.test_case_id} ---------------")
                timeout_resp = self.run_timeout
            elif "teardown" in fixture and "suite" in fixture:
                log.info(f"--------------- SUITE TEARDOWN: {self.test_suite_id} ---------------")
                timeout_resp = self.teardown_timeout
            elif "teardown" in fixture:
                log.info(f"--------------- TEARDOWN: {self.test_suite_id}, {self.test_case_id} ---------------")
                timeout_resp = self.teardown_timeout

            # for all configured auxiliaries just send the associated
            # command and evaluate return test results
            for aux in self.test_auxiliary_list:
                cmd = message.Message(
                    msg_type=message.MessageType.COMMAND,
                    sub_type=message_type,
                    test_suite=self.test_suite_id,
                    test_case=self.test_case_id,
                )
                is_ack = aux.send_fixture_command(command=cmd, timeout=timeout_cmd)

                # if the DUT has not acknowledge, just abort and try to
                # create a brand new communication stream with it
                if is_ack is False:
                    info_to_print = f"No response received from DUT for auxiliairy : {aux} command : {cmd}!"
                    self.cleanup_and_skip(aux, info_to_print)

                # wait until the DUT has run the required test fixture
                report = aux.wait_and_get_report(blocking=True, timeout_in_s=timeout_resp)

                # no report was received during the defined time slot,
                # just abort and try to create a brand new communication
                # stream
                if report is None:
                    info_to_print = (f"No report received from DUT for auxiliairy : {aux} command :{cmd}!",)
                    self.cleanup_and_skip(aux, info_to_print)

                is_test_on_dut_implemented = report.sub_type != message.MessageReportType.TEST_NOT_IMPLEMENTED
                is_report = report.get_message_type() == message.MessageType.REPORT

                # At least the test has to be implemented at embedded
                # side and a report type command was received
                if is_test_on_dut_implemented and is_report:
                    self.assertEqual(report.sub_type, message.MessageReportType.TEST_PASS)

            return func(self, *args, **kwargs)

        return handle_interaction

    return inner_interaction



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Virtual Can Communication Channel
*********************************

:module: cc_virtual_can

:synopsis: CChannel implementation for victual CAN using UDP multicast API from python-can

.. currentmodule:: cc_virtual_can

"""
import logging
from typing import Dict, Optional, Union

from pykiso import connector

try:
    import can
    from can.interfaces.udp_multicast.bus import UdpMulticastBus
except ImportError as e:
    raise ImportError(f"{e.name} dependency missing, consider installing pykiso with 'pip install pykiso[can]'")

log = logging.getLogger(__name__)


class CCVirtualCan(connector.CChannel):
    def __init__(
        self,
        channel: UdpMulticastBus = UdpMulticastBus.DEFAULT_GROUP_IPv4,
        receive_own_messages: bool = False,
        is_fd: bool = True,
        enable_brs: bool = False,
        is_extended_id: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.channel = channel
        self.interface = "udp_multicast"
        self.receive_own_messages = receive_own_messages
        self.is_fd = is_fd
        self.enable_brs = enable_brs
        self.is_extended_id = is_extended_id
        self._is_open = False
        self.timeout = 1e-6
        self.config = {
            "interface": self.interface,
            "channel": self.channel,
            "preserve_timestamps": False,
            "receive_own_messages": self.receive_own_messages,
            "fd": self.is_fd,
        }

    @property
    def is_open(self):
        return self._is_open

    def _cc_open(self) -> None:
        """Open a can bus channel."""
        if self.is_open:
            log.info(f"{self} is already open")
            return
        self.bus = can.Bus(**self.config)
        self._is_open = True

    def _cc_close(self) -> None:
        """Close the current can bus channel."""
        self.bus.shutdown()
        self.bus = None
        self._is_open = False

    def _cc_send(self, msg: bytes, remote_id: Optional[int] = None, **kwargs) -> None:
        """Send a CAN message at the configured id.

        If remote_id parameter is not given take configured ones

        :param msg: data to send
        :param remote_id: destination can id used
        :param kwargs: named arguments
        """

        remote_id = remote_id or self.remote_id

        can_msg = can.Message(
            arbitration_id=remote_id,
            data=msg,
            is_fd=self.is_fd,
            dlc=len(msg),
            is_extended_id=self.is_extended_id,
            bitrate_switch=self.enable_brs,
            check=True,
        )
        self.bus.send(can_msg)

        log.internal_debug("%s sent CAN Message: %s, data: %s", self, can_msg, msg)

    def _cc_receive(self, timeout: float = 0.0001) -> Dict[str, Union[bytes, int, None]]:
        """Receive a can message using configured filters.

        :param timeout: timeout applied on reception

        :return: the received data and the source can id
        """
        try:
            received_msg = self.bus.recv(timeout=timeout or self.timeout)

            if received_msg is not None:
                frame_id = received_msg.arbitration_id
                payload = received_msg.data
                timestamp = received_msg.timestamp

                log.internal_debug(
                    "received CAN Message: %s, %s, %s",
                    frame_id,
                    payload,
                    timestamp,
                )
                return {
                    "msg": payload,
                    "remote_id": frame_id,
                    "timestamp": timestamp,
                }
            else:
                return {"msg": None}
        except can.CanError as can_error:
            log.internal_info("encountered CAN error while receiving message: %s", can_error)
            return {"msg": None}
        except Exception:
            log.exception(f"encountered error while receiving message via {self}")
            return {"msg": None}



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Proxy auxiliary plugin
**********************

:module: proxy_auxiliary

:synopsis: implementation of existing ProxyAuxiliary for Robot
    framework usage.

.. currentmodule:: proxy_auxiliary

"""

from robot.api.deco import keyword, library

from ..auxiliaries.proxy_auxiliary import ProxyAuxiliary as ProxyAux
from .aux_interface import RobotAuxInterface


@library(version="0.1.0")
class ProxyAuxiliary(RobotAuxInterface):
    """Robot framework plugin for ProxyAuxiliary."""

    ROBOT_LIBRARY_SCOPE = "SUITE"

    def __init__(self):
        """Initialize attributes."""
        super().__init__(aux_type=ProxyAux)

    @keyword(name="Suspend")
    def suspend(self, aux_alias: str) -> None:
        """Suspend given auxiliary's run.

        :param aux_alias: auxiliary's alias
        """
        aux = self._get_aux(aux_alias)
        aux.suspend()

    @keyword(name="Resume")
    def resume(self, aux_alias: str) -> None:
        """Resume given auxiliary's run.

        :param aux_alias: auxiliary's alias
        """
        aux = self._get_aux(aux_alias)
        aux.resume()



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
UDS-specific Exceptions
***********************

:module: uds_exceptions

:synopsis: Exceptions to be raised in UDS

.. currentmodule:: uds_exceptions

"""


class UDSError(Exception):
    def __init__(self):
        self.msg = ""

    def __str__(self):
        return self.msg


class NegativeResponseError(UDSError):
    """Raised if UDS response is negative (starting with 0x7F)"""

    def __init__(self, response):
        self.msg = f"Request refused by client, got response: {response}"


class UnexpectedResponseError(UDSError):
    """Raised if UDS response is positive but should not be"""

    def __init__(self, response):
        self.msg = f"Request expected to fail but succeed, got response: {response}"


class InvalidResponseError(UDSError):
    """Raise if UDS response is not the expected one"""

    def __init__(self, command):
        self.msg = f"Invalid response for command {command}"


class ResponseNotReceivedError(UDSError):
    """Raise if no UDS response is received"""

    def __init__(self, command):
        self.msg = f"No response received for command {command}"



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
:module: exceptions

:synopsis: Define all custom exceptions raised by pykiso framework

.. currentmodule:: exceptions
"""


from typing import List, Union


class PykisoError(Exception):
    """Pykiso specific exception used as basis for all others."""

    def __str__(self):
        return self.message


class TestCollectionError(PykisoError):
    """Collection of test cases by the TestLoader failed."""

    __test__ = False

    def __init__(self, test_suite_dir: Union[str, List[str]]) -> None:
        """Initialize attributes.

        :param test_suite_dir: path to the test suite in which a test
            case failed to be loaded.
        """
        suites_dir = test_suite_dir if not isinstance(test_suite_dir, list) else ", ".join(test_suite_dir)
        self.message = f"Failed to collect test suites {suites_dir}"
        super().__init__(self.message)


class AuxiliaryCreationError(PykisoError):
    """Raised when an auxiliary instance creation failed."""

    def __init__(self, aux_name: str) -> None:
        """Initialize attributes.

        :param aux_name: configured auxiliary alias.
        """
        self.message = f"Auxiliary {aux_name} failed at instance creation"
        super().__init__(self.message)


class AuxiliaryNotStarted(PykisoError):
    """
    Raised when an action that requires an auxiliary to be running
    is executed although the auxiliary is stopped.
    """

    def __init__(self, aux_name: str) -> None:
        """Initialize attributes.

        :param aux_name: configured auxiliary alias.
        """
        self.message = f"Auxiliary {aux_name} was not started, cannot run command"
        super().__init__(self.message)


class ConnectorRequiredError(PykisoError):
    """Raised when an auxiliary instance creation failed because a connector is
    required and none is provided
    """

    def __init__(self, aux_name: str) -> None:
        """Initialize attributes.

        :param aux_name: configured auxiliary alias.
        """
        self.message = f"A connector is required for {aux_name} and none is provided"
        super().__init__(self.message)


class InvalidTestModuleName(PykisoError):
    """Raised when a invalid python module name is covered by test file pattern
    of either the cli or the yaml config
    """

    def __init__(self, name: str) -> None:
        self.message = (
            f"Test files need to be valid python module names but '{name}' is invalid"
            "\nModule name can only contain letters, numbers, _ (underscore), "
            "needs to start with a letter or underscore"
        )
        super().__init__(self.message)



##########################################################################
# Copyright (c) 2010-2023 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################
"""
ODX Parser for UDS Server Auxiliary
***********************************

:module: odx_parser

:synopsis: odx parser used by a uds server to dynamically configure its
    uds callbacks from an odx file

.. currentmodule:: odx_parser

"""
from __future__ import annotations

import logging
from enum import Enum
from pathlib import Path
from typing import List
from xml.etree.ElementTree import Element

# Use defusedxml, as xml is not secure against maliciously constructed data.
from defusedxml import ElementTree

log = logging.getLogger(__name__)


class OdxParser:
    """Used to parse ODX files to configure a Uds server"""

    class RefType(Enum):
        REQUEST = "REQUEST-REF"
        POS_RESPONSE = "POS-RESPONSE-REF"

    def __init__(self, odx_file: Path) -> None:
        with open(odx_file) as odx:
            # create a xml tree from root <ODX>
            self.odx_tree = ElementTree.parse(odx)

    def _find_element_by_odx_id(self, odx_id: str) -> Element:
        """Find an odx element by the given id

        :param id: odx id of the element to search for
        :raises ValueError: if no element with given odx id found
        :return: the element with the given odx id
        """
        # xpath finding the element with the given id
        element = self.odx_tree.find(f".//*[@ID='{odx_id}']")
        if element is None:
            raise ValueError(f"No element with id={odx_id} found")
        return element

    def _find_diag_services_by_sd(self, sd_instance_name: str) -> Element:
        """Find the <DIAG-SERVICE> with the given sd_instance_name

        :param sd_instance_name: content of the <SD SI="DiagInstanceName">
        :return: the parent diag service odx element of the sd element with given name
        """
        # xpath to diag service for given sd instance name
        diag_services = self.odx_tree.findall(f".//SD[@SI][.='{sd_instance_name}']../../..")
        if not diag_services:
            raise ValueError(f"No DIAG-SERVICE has a SD containing {sd_instance_name}")
        return diag_services

    def get_coded_values(self, sd: str, sid: int, ref_type: RefType = RefType.REQUEST) -> List[int]:
        """Get the list of coded values for a ODX request element to construct a UDS request from

        :param sd: sd instance name
        :param sid: service id to differentiate between diag services with the same sd name
        :raises ValueError: if no request could be created for the given sd name and SID
        :return: a list of the coded values to be converted into a uds request
        """
        diag_services = self._find_diag_services_by_sd(sd)
        for diag_service in diag_services:
            request_id = diag_service.find(f".//{ref_type.value}").attrib["ID-REF"]
            request_element = self._find_element_by_odx_id(request_id)
            # compare SIDs to differentiate between e.g. read and write request with same sd name
            request_sid = int(request_element.find(".//PARAM[@SEMANTIC='SERVICE-ID']/CODED-VALUE").text)
            if request_sid == sid:
                coded_value_elements = request_element.findall(".//CODED-VALUE")
                coded_values = [int(coded_value.text) for coded_value in coded_value_elements]
                return coded_values
        log.error(f"Could not create request for service={sid} and sd={sd}")
        raise ValueError(f"Could not create request for service={sid} and sd={sd}")



##########################################################################
# Copyright (c) 2010-2023 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Report customization
********************

:module: reporting

:synopsis: customize the generated reports with pykiso-specific properties.

"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .utils import *

if TYPE_CHECKING:
    from _pytest.nodes import Item
    from _pytest.runner import CallInfo

    from pykiso.test_coordinator.test_case import BasicTest


@export
def pytest_runtest_makereport(item: Item, call: CallInfo[None]) -> None:
    """Add the test IDs to the JUnit report to the user properties in order to add
    them to the generated reports.

    :param item: collected test item.
    :param call: the result of the test item's execution.
    """
    if call.when != "setup":
        return

    if is_kiso_testcase(item):
        kiso_tc: BasicTest = get_base_testcase(item)
        test_ids = kiso_tc.test_ids
    else:
        test_ids_marker = item.get_closest_marker("test_ids", default=None)
        test_ids = test_ids_marker.kwargs if test_ids_marker is not None else None

    item.user_properties.append(("test_ids", test_ids))



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Collection of ready to use auxiliaries
**************************************

:module: auxiliaries

.. currentmodule:: auxiliaries

"""



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Can Communication Channel SocketCAN
***********************************

:module: cc_socket_can

:synopsis: CChannel implementation for CAN(fd) using SocketCAN

.. currentmodule:: cc_socket_can

"""

import logging
import platform
import time
from pathlib import Path
from typing import Dict, Literal, Optional, Union

try:
    import can
    import can.bus
except ImportError as e:
    raise ImportError(f"{e.name} dependency missing, consider installing pykiso with 'pip install pykiso[can]'")


from pykiso import CChannel, Message
from pykiso.lib.connectors.cc_socket_can.socketcan_to_trc import SocketCan2Trc, can

MessageType = Union[Message, bytes]

log = logging.getLogger(__name__)


def os_name() -> str:
    """Returns the system/OS name.

    :return: os name such as 'Linux', 'Darwin', 'Java', 'Windows'
    """
    return platform.system()


class CCSocketCan(CChannel):
    """CAN FD channel-adapter."""

    def __init__(
        self,
        channel: str = "vcan0",
        remote_id: Optional[int] = None,
        is_fd: bool = True,
        enable_brs: bool = False,
        can_filters: Optional[list] = None,
        is_extended_id: bool = False,
        receive_own_messages: bool = False,
        logging_activated: bool = False,
        trace_path: Optional[str] = None,
        log_name: Optional[str] = None,
        strategy_trc_file: Optional[Literal["testRun", "testCase"]] = None,
        **kwargs,
    ):
        """Initialize can channel settings.

        :param channel: the can interface name. (i.e. vcan0, can1, ..)
        :param remote_id: id used for transmission
        :param is_fd: should the Bus be initialized in CAN-FD mode
        :param enable_brs: sets the bitrate_switch flag to use higher transmission speed
        :param can_filters: iterable used to filter can id on reception
        :param is_extended_id: this flag controls the size of the arbitration_id field
        :param receive_own_messages: if set transmitted messages will be received
        :param logging_activated: boolean used to enable logfile creation
        :param trace_path: trace directory path (absolute or relative), if you set a trc file name, auto timestamp and log_name will be ignored
        :param log_name: trace full name (without file extension)
        """
        super().__init__(**kwargs)
        self.channel = channel
        self.remote_id = remote_id
        self.is_fd = is_fd
        self.enable_brs = enable_brs
        self.can_filters = can_filters if can_filters else []
        self.is_extended_id = is_extended_id
        self.receive_own_messages = receive_own_messages
        self.logging_activated = logging_activated
        self.bus = None
        self.logger = None
        self.trace_path = trace_path
        self.log_name = log_name.strip(".trc") if log_name else None
        self.strategy_trc_file = strategy_trc_file
        self.opened = False

        # Set a timeout to send the signal to the GIL to change thread.
        # In case of a multi-threading system, all tasks will be called one after the other.
        self.timeout = 1e-6

        if self.logging_activated:
            self.resolve_trace_path()
        if self.enable_brs and not self.is_fd:
            log.internal_warning("Bitrate switch will have no effect because option is_fd is set to false.")

    def resolve_trace_path(self):
        """Resolve the trace path and name of the file that will be created for the trace"""
        # if no trace path is given, take the root path (where pykiso is launched)
        if self.trace_path is None or self.trace_path == "":
            self.trace_path = Path().resolve()

        # if the given log path is not absolute add root path
        # (where pykiso is launched) otherwise take it as it is
        self.trace_path = (
            Path().resolve() / self.trace_path if not Path(self.trace_path).is_absolute() else Path(self.trace_path)
        )

        # if specify a trace path with a file name, take the parent directory, ignore log_name and use the trc file as given.
        if self.trace_path.suffix == ".trc":
            self.log_name = self.trace_path.with_suffix("").name
            self.trace_path = self.trace_path.parent
        else:  # add timestamp to the log name other wise use timestamp with default name
            # add timestamp to the log name other wise use timestamp with default name
            self.log_name = (
                time.strftime(f"%Y-%m-%d_%H-%M-%S_{self.log_name}")
                if self.log_name is not None
                else time.strftime("%Y-%m-%d_%H-%M-%S_CanLog")
            )

        # create the trace path if it does not exist
        self.trace_path.mkdir(parents=True, exist_ok=True)
        self.log_name = f"{self.log_name}.trc"

    def _cc_open(self) -> None:
        """Open a can bus channel, set filters for reception and activate"""
        if self.opened:
            log.warning("Socket can is already opened")
            return
        if not os_name() == "Linux":
            raise OSError("socketCAN is only available under linux.")

        self.bus = can.interface.Bus(
            interface="socketcan",
            channel=self.channel,
            can_filters=self.can_filters,
            fd=self.is_fd,
            receive_own_messages=self.receive_own_messages,
        )

        if self.logging_activated:
            log.internal_info(f"Logging path for socketCAN set to {self.trace_path / self.log_name} ")
            self.logger = SocketCan2Trc(self.channel, str(self.trace_path / self.log_name))
            if self.strategy_trc_file is None:
                self.logger.start()
        self.opened = True

    def _cc_close(self) -> None:
        """Close the current can bus channel and close the log handler."""
        if not self.opened:
            log.warning("Socket can is already closed")
            return
        self.bus.shutdown()
        self.bus = None

        if self.logging_activated:
            self.logger.stop()
            self.logger = None

        self.opened = False

    def _cc_send(self, msg: MessageType, remote_id: int | None = None, **kwargs) -> None:
        """Send a CAN message at the configured id.
        If remote_id parameter is not given take configured ones

        :param msg: data to send
        :param remote_id: destination can id used

        """
        remote_id = remote_id or self.remote_id

        can_msg = can.Message(
            arbitration_id=remote_id,
            data=msg,
            is_extended_id=self.is_extended_id,
            is_fd=self.is_fd,
            bitrate_switch=self.enable_brs,
        )
        self.bus.send(can_msg)

        log.internal_debug(f"{self} sent CAN Message: {can_msg}, data: {msg}")

    def _cc_receive(self, timeout: float = 0.0001) -> Dict[str, Union[bytes, int]]:
        """Receive a can message using configured filters.

        :param timeout: timeout applied on reception

        :return: tuple containing the received data and the source can id
        """
        try:  # Catch bus errors & rcv.data errors when no messages where received
            received_msg = self.bus.recv(timeout=timeout or self.timeout)
            if received_msg is not None:
                frame_id = received_msg.arbitration_id
                payload = received_msg.data
                timestamp = received_msg.timestamp

                log.internal_debug("received CAN Message: {}, {}, {}".format(frame_id, payload, timestamp))
                return {"msg": payload, "remote_id": frame_id}
            else:
                return {"msg": None}
        except can.CanError as can_error:
            log.internal_debug(f"encountered can error: {can_error}")
            return {"msg": None}
        except Exception:
            log.exception(f"encountered error while receiving message via {self}")
            return {"msg": None}

    def stop_can_trace(self):
        """
        Stops the CAN trace if it is currently running.
        :return: None
        """
        if not self.logging_activated or self.logger is None:
            return

        self.logger.stop()

    def start_can_trace(self, trace_path: Optional[str] = None, log_name: Optional[str] = None) -> None:
        """Start the trace and write to the trace path written or the log name.

        :param trace_path: trace directory path (absolute or relative), if you set a trc file name, auto timestamp and log_name will be ignored,
            if not defined, will keep the directory path given in the configuration
        :param log_name: trace full name (without file extension)
        """
        if not self.logging_activated or self.logger is None:
            return

        if not self.logger.started:
            self.log_name = log_name.strip(".trc") if log_name else None
            if trace_path is not None:
                self.trace_path = trace_path
                self.resolve_trace_path()
            log.internal_info(f"Logging path for socketCAN set to {self.trace_path / self.log_name} ")
            self.logger.trc_file_name = str(self.trace_path / self.log_name)
            self.logger.start()



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Auxiliary Interface Definition
******************************

:module: dynamic_loader

:synopsis: Import magic that enables aliased auxiliary loading in TestCases


"""

from __future__ import annotations

import importlib
import importlib.abc
import importlib.machinery
import importlib.util
import logging
import pathlib
import sys
import types
from typing import TYPE_CHECKING, Type, Union

from pykiso.exceptions import ConnectorRequiredError

if TYPE_CHECKING:
    from ..auxiliary import AuxiliaryInterface
    from ..connector import Connector

PACKAGE = __package__.split(".")[0]

__all__ = ["DynamicImportLinker"]

log = logging.getLogger(__name__)


class DynamicFinder(importlib.abc.MetaPathFinder):
    """A MetaPathFinder that delegates everything to the loader."""

    def __init__(self, loader: AuxLinkLoader):
        """Initialize attributes.

        :param loader: any Loader object from importlib.abc.Loader
        """
        self._loader = loader

    def find_spec(self, fullname, path, target=None):
        """Attempt to locate the requested module.

        :param fullname: is the fully-qualified name of the module,
        :param path: is set to __path__ for sub-modules/packages,
            or None otherwise.
        :param target: can be a module object, but is unused in this example.
        """
        if self._loader.provides(fullname):
            return self._gen_spec(fullname)

    def _gen_spec(self, fullname):
        spec = importlib.machinery.ModuleSpec(fullname, self._loader)
        return spec


class AuxLinkLoader(importlib.abc.Loader):
    """A Loader for auxiliaries.

    Something is imported from `pykiso.auxiliaries` gets redirected to
    a lookup table of configured auxiliaries (done via config file).
    """

    _COMMON_PREFIX = PACKAGE + ".auxiliaries"

    def __init__(self, aux_cache: AuxiliaryCache):
        """Initialize attributes.

        :param aux_cache: AuxiliaryChache instance
        """
        self._auxiliaries = []
        self._aux_cache = aux_cache
        # create a dummy module to return when Python attempts to import
        # myapp and myapp.virtual, the :-1 removes the last "." for
        # aesthetic reasons :)
        self._dummy_module = types.ModuleType(self._COMMON_PREFIX.rstrip("."))
        # set __path__ so Python believes our dummy module is a package
        # this is important, since otherwise Python will believe our
        # dummy module can have no submodules
        self._dummy_module.__path__ = []

    def provide(self, name: str):
        """Register a service as provided via the given module.

        A service is any Python object in this context - an imported module,
        a class, etc.

        :param name: aliased auxiliary instance
        """
        self._auxiliaries.append(name)

    def provides(self, fullname: str) -> bool:
        """Check if this loader provides the requested module.

        :param fullname: fully qualified name (e.g. pykiso.auxiliaries.alias)
        """
        if self._truncate_name(fullname) in self._auxiliaries:
            return True
        else:
            # this checks if we should return the dummy module,
            # since this evaluates to True when importing myapp and
            # myapp.virtual
            return self._COMMON_PREFIX.startswith(fullname)

    def create_module(self, spec: importlib.machinery.ModuleSpec) -> Union[types.ModuleType, AuxiliaryInterface]:
        """Create the given module from the supplied module spec.

        Under the hood, this module returns a service or a dummy module,
        depending on whether Python is still importing one of the names listed
        in _COMMON_PREFIX.
        """
        name = self._truncate_name(spec.name)
        if name not in self._auxiliaries:
            # return our dummy module since at this point we're loading
            # *something* along the lines of "pykiso.components" that's not
            # a module
            return self._dummy_module
        return self._aux_cache.get_instance(name)

    def exec_module(self, module):
        """Execute the given module in its own namespace.

        This method is required to be present by importlib.abc.Loader,
        but since we know our module object is already fully-formed,
        this method merely no-ops.
        """
        pass

    def _truncate_name(self, fullname: str) -> str:
        """Strip off _COMMON_PREFIX from the given module name.

        Convenience method when checking if a service is provided.
        """
        return fullname[len(self._COMMON_PREFIX) + 1 :]


class ModuleCache:
    """Caches modules, configs and instances.

    This class serves as a cache for configurations and instances for modules to be loaded.

    An entry consists of

    * an alias for the instance
    * a location (path to python file or python module)
    * a class
    * configuration parameters
    * active instance (lazily created)
    """

    def __init__(self):
        """Initialize attributes."""
        self.locations = dict()
        self.configs = dict()
        self.modules = dict()
        self.instances = dict()

    def provide(self, name: str, module: str, **config_params):
        """Provide an aliased instance.

        :param name: the instance alias
        :param module: either 'python/file/path.py:Class' or 'module:Class'
            of the class we want to provide
        """
        self.locations[name] = module
        self.configs[name] = config_params

    def _import(self, name: str) -> Type[Union[AuxiliaryInterface, Connector]]:
        """Import the class registered under the alias <name>."""
        try:
            import_path = self.locations[name]
            location, _class = import_path.rsplit(":", 1)
        except KeyError:
            raise ValueError(f"Could not find {name!r} in provided configuration")
        except ValueError:
            raise ValueError(f"Specified type for {name!r} must be 'path:Class' or 'module:Class', got {import_path!r}")
        if ".py" in location:
            path_loc = pathlib.Path(location)
            if path_loc.exists() and path_loc.is_file():
                spec = importlib.util.spec_from_file_location(path_loc.stem, path_loc)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                log.internal_debug(f"loading {_class} as {name} from {path_loc}")
            else:
                raise ImportError(f"no python module found at {path_loc!r}", name=_class)
        else:
            module = importlib.import_module(location)
        cls = getattr(module, _class)
        log.internal_debug(f"loaded {_class} as {name} from {location}")
        return cls

    def get_instance(self, name: str) -> Union[AuxiliaryInterface, Connector]:
        """Get an instance of alias <name> (create and configure one of not existed)."""
        if name in self.instances:
            log.internal_debug(f"instance for {name} found ({self.instances[name]})")
            return self.instances[name]
        if name not in self.modules:
            log.internal_debug(f"module for {name} not found, loading...")
            self.modules[name] = self._import(name)
        log.internal_debug(f"instantiating {name}: {self.modules[name]}({self.configs[name]})")
        inst = self.modules[name](name=name, **self.configs[name])
        self.instances[name] = inst
        log.internal_debug(f"instantiated {name}")
        return inst

    def delete_all_instances(self) -> None:
        """Call shutdown method if it exists"""
        for instance in self.instances.values():
            if callable(getattr(instance, "shutdown", None)):
                instance.shutdown()


class AuxiliaryCache(ModuleCache):
    """A ModuleCache that specifically provides Auxiliaries.

    This has the additional functionality that if an auxiliary has any defined connectors,
    these will be provided automatically.
    """

    def __init__(self, con_cache: ModuleCache):
        """Initialize attributes.

        :param con_cache: connector ModuleCache object
        """
        super().__init__()
        self.con_cache = con_cache
        self.connectors = dict()

    def provide(self, name: str, module: str, connectors=None, **config_params):
        """Provide an aliased instance.

        :param name: the instance alias
        :param module: either 'python-file-path:Class' or 'module:Class' of the class we want
        :param connectors: list of connector aliases
            to provide
        """
        self.connectors[name] = connectors
        super().provide(name, module, **config_params)

    def get_instance(self, name: str) -> AuxiliaryInterface:
        """Get an instance of alias <name> (create and configure one of not existed)."""
        for cn, con in self.connectors.get(name, dict()).items():
            # add connector-instances as configs
            self.configs[name][cn] = self.con_cache.get_instance(con)
        inst = super().get_instance(name)

        if getattr(inst, "connector_required", True) and not getattr(inst, "channel", False):
            self.instances.pop(name)
            raise ConnectorRequiredError(name)
        # if auto start is needed start the auxiliary otherwise store
        # the created instance
        auto_start = getattr(inst, "auto_start", True)

        if not inst.is_instance and auto_start:
            inst.start()
            inst.create_instance()
            # Can't used the CChannel typing here due to cyclic import
            if hasattr(inst, "channel") and inst.channel is not None and inst.channel.auto_open:
                inst.channel.open()
            log.internal_debug(f"called create_instance on {name}")
        self.instances[name] = inst
        return inst

    def _stop_auxiliaries(self):
        """Elegant workaround to shut down all the auxiliaries."""
        for alias, aux in self.instances.items():
            log.internal_debug(f"issuing stop for auxiliary '{aux}'")
            aux.stop()
            aux_mod = f"{AuxLinkLoader._COMMON_PREFIX}.{alias}"
            # ensure that the module was created
            if sys.modules.get(aux_mod) is not None:
                # remove all modules create by our custom loader
                sys.modules.pop(aux_mod)

        # remove the common prefix "pykiso.auxiliaries" to enforce the
        # path finder -> loader -> module
        if AuxLinkLoader._COMMON_PREFIX in sys.modules:
            sys.modules.pop(AuxLinkLoader._COMMON_PREFIX)


class DynamicImportLinker:
    """Public Interface of Import Magic.

    initialises the Loaders, Finders and Caches, implements interfaces to
    install the magic and register the auxiliaries and connectors."""

    def __init__(self):
        """Initialize attributes."""
        self._con_cache = ModuleCache()
        self._aux_cache = AuxiliaryCache(self._con_cache)
        self._aux_loader = AuxLinkLoader(self._aux_cache)
        self._finders = [DynamicFinder(self._aux_loader)]

    def install(self):
        """Install the import hooks with the system."""
        log.internal_debug(f"installed the {self.__class__.__name__}")
        sys.meta_path.insert(0, *self._finders)

    def provide_connector(self, name: str, module: str, **config_params):
        """Provide a connector.

        :param name: the connector alias
        :param module: either 'python-file-path:Class' or 'module:Class'
            of the class we want to provide.
        """
        log.internal_debug(f"provided connector {name} (at {module})")
        self._con_cache.provide(name, module, **config_params)

    def provide_auxiliary(self, name: str, module: str, aux_cons=None, **config_params):
        """Provide a auxiliary.

        :param name: the auxiliary alias
        :param module: either 'python-file-path:Class' or 'module:Class'
            of the class we want to provide.
        :param aux_cons: list of connectors this auxiliary has
        """
        log.internal_debug(f"provided auxiliary {name} (at {module})")
        self._aux_cache.provide(name, module, connectors=aux_cons, **config_params)
        self._aux_loader.provide(name)

    def uninstall(self):
        """Deregister the import hooks, close all running threads, delete all instances."""
        log.internal_debug("closing and uninstalling all dynamic modules and loaders")
        self._stop_auxiliaries()
        self._con_cache.delete_all_instances()
        del self._con_cache
        del self._aux_cache
        del self._aux_loader
        for finder in self._finders:
            sys.meta_path.remove(finder)

    def _stop_auxiliaries(self):
        """Elegant workaround to shut down all the auxiliaries."""
        self._aux_cache._stop_auxiliaries()



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Communication Channel Via segger j-link
***************************************

:module: cc_rtt_segger

:synopsis: channel used to enable RTT communication using Segger J-Link debugger.
    Additionally, RTT logs can be captured by setting the rtt_log_path parameter
    on the specified channel.

.. currentmodule:: cc_rtt_segger

"""
import functools
import logging
import threading
import time
from pathlib import Path
from typing import Dict, Optional

try:
    import pylink
    from pylink.errors import JLinkRTTException
except ImportError as e:
    raise ImportError(f"{e.name} dependency missing, consider installing pykiso with 'pip install pykiso[debugger]'")


from pykiso import connector
from pykiso.message import Message

log = logging.getLogger(__name__)


def _need_connection(f):
    """Decorator to check the JLink is connected and raises an error otherwise"""

    @functools.wraps(f)
    def check_before_execution(self, *args, **kwargs):
        """Check if Jlink is opened, else raise error

        :raises pylink.JLinkException: if Jlink not available
        """
        if not self.jlink.opened():
            raise pylink.JLinkException("You need to connect to a JLink first")
        return f(self, *args, **kwargs)

    return check_before_execution


def _need_rtt(f):
    """Decorator to check that a RTT connection has been configured and raises an error otherwise"""

    @functools.wraps(f)
    def check_before_execution(self, *args, **kwargs):
        """Check if Jlink is opened and RTT configured, else raise errors

        :raises pylink.JLinkException: if RTT config not done or Jlink not available
        """

        if not self.jlink.opened():
            raise pylink.JLinkException("You need to connect to a JLink first")
        if not self.rtt_configured:
            raise pylink.JLinkException("You need to configure RTT first")
        return f(self, *args, **kwargs)

    return check_before_execution


class CCRttSegger(connector.CChannel):
    """Channel using RTT to communicate through Segger J-Link debugger."""

    def __init__(
        self,
        serial_number: int = None,
        chip_name: str = "STM32L562QE",
        speed: int = 4000,
        block_address: int = 0x2003F800,
        verbose: bool = False,
        tx_buffer_idx: int = 3,
        rx_buffer_idx: int = 0,
        rtt_log_path: Optional[str] = None,
        rtt_log_buffer_idx: int = 0,
        rtt_log_speed: float = 1000,
        rtt_decode_strategy: str = "replace",
        connection_timeout: int = 5,
        search_range: int = None,
        **kwargs,
    ):
        """Initialize attributes.

        :param serial_number: optional segger debugger serial number (required if many connected)
        :param chip_name: microcontoller name (STM....)
        :param speed: communication speed in Hz
        :param block_address: start address to start RTT communication
        :param tx_buffer_idx: buffer index used for transmission
        :param rx_buffer_idx: buffer index used for reception
        :param verbose: boolean indicating if J-Link connection should be verbose in logging
        :param rtt_log_path: path to the folder where the RTT log file should be stored
        :param rtt_log_buffer_idx: buffer index used for RTT logging
        :param rtt_log_speed: number of log per second to be pulled (manage the CPU load for logging)
            None value fetch log at the CPU's speed. Default 1000 logs/s
        :param rtt_decode_strategy: how to handle undecodable bytes received on the rtt stream.
            Possible options are 'ignore' or 'replace' which replace the character with an suitable replacement
            character. Default 'replace'
        :param connection_timeout: available time (in seconds) to open the connection
        :param search_range: range length to search for the block of the device
            starting from the specified block address.
        """
        super().__init__(**kwargs)
        self.serial_number = serial_number if isinstance(serial_number, int) else None
        self.chip_name = chip_name
        self.speed = speed
        self.block_address = block_address
        self.tx_buffer_idx = tx_buffer_idx
        self.rx_buffer_idx = rx_buffer_idx
        self.verbose = verbose
        self.jlink = None
        self.connection_timeout = connection_timeout
        self.rtt_log_buffer_idx = rtt_log_buffer_idx
        self.rx_buffer_size = None
        # initialize rtt logging specific parameters
        self.rtt_configured = False
        self._log_thread_running = False
        self.rtt_log_refresh_time = round(1 / rtt_log_speed, 6) if rtt_log_speed else 0
        self.rtt_log_thread = threading.Thread(target=self.receive_log)
        self.rtt_log_path = rtt_log_path
        self.rtt_log = logging.getLogger(f"{__name__}{serial_number or ''}.RTT")
        self.search_range = search_range

        if rtt_decode_strategy in ["replace", "ignore"]:
            self.rtt_decode_strategy = rtt_decode_strategy
        else:
            raise ValueError("rtt_decode_strategy must be 'replace' or 'ignore'")

        if self.rtt_log_path is not None:
            self.rtt_log_buffer_size = 0
            self.rtt_log_path = Path(rtt_log_path)
            # if log folder does not exists create it during process
            self.rtt_log_path.mkdir(parents=True, exist_ok=True)
            rtt_fh = logging.FileHandler(self.rtt_log_path / "rtt.log")
            rtt_fh.setLevel(logging.DEBUG)
            self.rtt_log.setLevel(logging.DEBUG)
            self.rtt_log.addHandler(rtt_fh)

    def read_target_memory(self, addr: int, num_units: int, zone: str = None, nbits: int = 32) -> Optional[list]:
        """Read the given target's memory units and the given address.

        .. note:: The optional ``zone`` specifies a memory zone to
            access to read from, e.g. ``IDATA``, ``DDATA``, or ``CODE``.

        .. warning:: The given number of bits, if provided, must be
            either ``8``, ``16``, or ``32``.  If not provided, always
            reads 32 bits.

        :param addr: start address to read from
        :param num_units: number of units to read
        :param zone: optional memory zone name to access
        :param nbits: number of bits to use for each unit

        :return: List of units read from the target.
        """
        mem_vals = None

        try:
            mem_vals = self.jlink.memory_read(addr, num_units, zone, nbits)
        except pylink.errors.JLinkException:
            log.exception(f"encountered error while reading memory at {hex(addr)}")
        except ValueError:
            log.exception("wrong number of bits given : must be 8, 16 or 32 bits")

        return mem_vals

    def _cc_open(self) -> None:
        """Connect debugger/microcontroller and retry if the connection fails due
        to instability
        """
        try:
            self._connect_jlink()
        except (JLinkRTTException, ValueError):
            log.error("An error occurred while connecting to the Jlink, trying a second time")
            self._cc_close()
            self._connect_jlink()

    def _connect_jlink(self) -> None:
        """Connect debugger/microcontroller.

        This method proceed to the following actions :
        - create a JLink class instance
        - connect to  the debugger(using open method)
        - set debugger interface to SWD
        - connect debugger to the specified chip
        - start RTT communication
        - start RTT Logging the specified channel if activated

        :raise JLinkRTTException: if connection timeout occurred.
        """
        self.jlink = pylink.JLink()

        # connect to J-Link debugger
        if not self.jlink.opened():
            self.jlink.open(self.serial_number)
            log.internal_info(f"connection made with J-Link debugger {self.serial_number}")
        else:
            log.internal_debug("connection to J-Link already started")

        if self.search_range:
            self.jlink.exec_command(f"SetRTTSearchRanges {hex(self.block_address)} {hex(self.search_range)}")
            log.internal_info(
                f"RTT Search range has been set at the following address {hex(self.block_address)} and size {hex(self.search_range)}"
            )

        # set target interface to SWD
        self.jlink.set_tif(pylink.enums.JLinkInterfaces.SWD)
        # connect debugger to  the specified target
        self.jlink.connect(chip_name=self.chip_name, speed=self.speed, verbose=self.verbose)
        log.internal_debug(f"connection to chip {self.chip_name} performed at speed {self.speed}Hz")
        # reset debugger if halted
        if self.jlink.halted():
            log.internal_info(f"J-Link is halted, reset target and wait for {self.connection_timeout}s")
            self.jlink.reset(halt=False)
            time.sleep(self.connection_timeout)
        # start rtt at the specified address
        self.jlink.rtt_start(self.block_address)
        if self.block_address:
            log.internal_info(f"RTT communication started at address {hex(self.block_address)}")
        t_start = time.perf_counter()
        while True:
            try:
                num_up = self.jlink.rtt_get_num_up_buffers()
                num_down = self.jlink.rtt_get_num_down_buffers()
                log.internal_debug(f"RTT started. Found {num_up} up and {num_down} down channels.")
                # get rx buffer size
                rx_buffer = self.jlink.rtt_get_buf_descriptor(self.rx_buffer_idx, True)
                self.rx_buffer_size = rx_buffer.SizeOfBuffer
                log.internal_debug(f"Maximum size for a received message set to {self.rx_buffer_size}")
                self.rtt_configured = True
                break
            except pylink.errors.JLinkRTTException:
                time.sleep(0.1)
                # Exit the while loop once timeout is reached
                if time.perf_counter() > (t_start + self.connection_timeout):
                    raise

        # start rtt logging thread on buffer index rtt_log_buffer_idx
        if self.rtt_log_path is not None:
            try:
                rtt_log_buffer = self.jlink.rtt_get_buf_descriptor(self.rtt_log_buffer_idx, True)
                self.rtt_log_buffer_size = rtt_log_buffer.SizeOfBuffer
                if self.rtt_log_buffer_size == 0:
                    raise ValueError
                log.internal_debug(f"RTT log buffer size is {self.rtt_log_buffer_size} bytes")
            except ValueError:
                log.internal_debug("Read RTT log buffer size is 0, defaulting to 1kB")
                self.rtt_log_buffer_size = 1024
            except pylink.errors.JLinkRTTException as e:
                log.error(f"Could not get RTT log buffer size: {e}")
                if self.rtt_log_buffer_idx not in range(num_up + 1):
                    log.error(f"Buffer index must be at most {num_up}")
                    self.rtt_log_buffer_idx = 0
                self.rtt_log_buffer_size = 1024
            finally:
                self._log_thread_running = True
                self.rtt_log_thread.start()
                log.internal_info("RTT logging started")

    def _cc_close(self) -> None:
        """Close current RTT communication in use."""

        if self.jlink is not None:
            if self._log_thread_running:
                self._log_thread_running = False
                self.rtt_log_thread.join()
            self.jlink.rtt_stop()
            self.jlink.close()
            log.internal_info("RTT communication closed")

    def _cc_send(self, msg: bytes, **kwargs) -> None:
        """Send message using the corresponding RTT buffer.

        :param msg: message to send, should be bytes.
        :param kwargs: not used
        """
        try:
            msg = list(msg)
            bytes_written = self.jlink.rtt_write(self.tx_buffer_idx, msg)
            log.internal_debug(
                "===> message sent (RTT) on buffer %d: %s, number of bytes written: %d",
                self.tx_buffer_idx,
                msg,
                bytes_written,
            )
        except Exception:
            log.exception(f"ERROR occurred while sending {len(msg)} bytes on buffer {self.tx_buffer_idx}")

    def _cc_receive(self, timeout: float = 0.1, size: int = None, **kwargs) -> Dict[str, Optional[bytes]]:
        """Read message from the corresponding RTT buffer.

        :param timeout: timeout applied on receive event
        :param size: maximum amount of bytes to read
        :return: dictionary containing the received bytes if successful, otherwise None
        """
        is_timeout = False
        # maximum amount of bytes to read out
        size = size or self.rx_buffer_size
        t_start = time.perf_counter()

        # rtt_read is not a blocking method due to this fact a while loop is used
        # to act like a blocking ones.
        while not is_timeout:
            try:
                # Read the message header or all of the buffer
                msg_received = self.jlink.rtt_read(self.rx_buffer_idx, size)
                # if a message is received
                if msg_received:
                    # Parse the bytes list into bytes string
                    msg_received = bytes(msg_received)
                    log.internal_debug(
                        "<=== message received (RTT) on buffer %d: %s, number of bytes read: %d",
                        self.rx_buffer_idx,
                        msg_received,
                        len(msg_received),
                    )
                    break
            except Exception:
                log.exception(f"encountered error while receiving message via {self} on buffer {self.rx_buffer_idx}")
                return {"msg": None}

            # Exit the while loop once timeout is reached
            if time.perf_counter() > (t_start + timeout):
                is_timeout = True
                msg_received = None

        return {"msg": msg_received}

    @_need_rtt
    def receive_log(self) -> None:
        """Receive RTT log messages from the corresponding RTT buffer."""
        while self._log_thread_running:
            # receive at most rtt_log_buffer_size of RTT logs
            log_msg = self.jlink.rtt_read(self.rtt_log_buffer_idx, self.rtt_log_buffer_size)
            if log_msg:
                self.rtt_log.debug(bytes(log_msg).decode("utf8", errors=self.rtt_decode_strategy))
            time.sleep(self.rtt_log_refresh_time)  # reduce resource consumption

    @_need_connection
    def reset_target(self, wait_time: int = 100, halt: bool = False) -> None:
        """Reset target via JLink.

        :param wait_time: Amount of milliseconds to delay after reset
        :param halt: if the CPU should halt after reset
        """
        logging.info("Reset Target")
        self.jlink.enable_reset_pulls_reset()
        self.jlink.reset(wait_time, halt)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Library of SCPI commands
************************

:module: lib_scpi_commands

:synopsis: Library of helper functions used to send requests to
    instruments with SCPI protocol. This library can be used with any
    VISA instance having a write and a query method.

.. currentmodule:: lib_scpi_commands

"""
import logging
from typing import Tuple

from .lib_instruments import REGISTERED_INSTRUMENTS, SCPI_COMMANDS_DICT

log = logging.getLogger(__name__)


class LibSCPI:
    """Class containing common SCPI commands for write and query
    requests.
    """

    def __init__(self, visa_object, instrument: str = ""):
        """Constructor.

        :param visa_object: any visa object having a write and a query
            method
        :param instrument: name of the instrument in use. If registered,
             the commands adapted to this instrument's capabilities are
             used instead of the default ones.
        """
        self.visa_object = visa_object
        # Check if the instrument in use is registered
        if instrument in REGISTERED_INSTRUMENTS:
            self.instrument = instrument
            log.internal_info(f"{instrument} is a registered instrument: using registered commands")
        else:
            self.instrument = "DEFAULT"
            log.internal_info("Working with unknown instrument: using default commands.")

    def _send_scpi_command(
        self,
        cmd_tag: str,
        cmd_type: str,
        cmd_payload: str = "",
        cmd_validation: tuple = None,
    ) -> str:
        """Send the appropriate command to send to the instrument in
        use. Check if a command is different or not available for a
        registered instrument, uses the default command otherwise. If a
        command is registered to be unavailable for the instrument in
        use, the command is not sent, a warning is raised, and an empty
        respond message is returned.

        :param cmd_type: either 'write' or 'query'
        :param cmd_tag: command tag corresponding to the command to
            execute
        :param cmd_payload: payload of the command to execute
            (only used in write commands)
        :param cmd_validation: expected output after validation
            (only used in write commands)

        :return: the response message of the executed command
            (if query), the success code of the command (if write), or
            an empty message if the command to execute is unsupported by
            the instrument in use.
        """
        # Retrieving command to send

        command, validation = self.get_command(cmd_tag, cmd_type, cmd_validation)

        if command == "COMMAND_NOT_AVAILABLE":
            return command

        if cmd_type == "query":
            return self.visa_object.query(command)

        return self.visa_object.write(f"{command} {cmd_payload}".strip(), validation)

    def get_command(self, cmd_tag: str, cmd_type: str, cmd_validation: tuple = None) -> Tuple:
        """Return the pre-defined command.

        :param cmd_tag: command tag corresponding to the command to
            execute
        :param cmd_type: either 'write' or 'query'
        :param cmd_validation: expected output after validation
            (only used in write commands)

        :return: the associated command plus a tuple containing the
            associated query and the expected response (if
            cmd_validation is not none) otherwise None
        """
        validation = None
        # Retrieving command to send
        if self.instrument in SCPI_COMMANDS_DICT[cmd_tag][cmd_type]:
            instr = self.instrument
        else:
            instr = "DEFAULT"
        command = SCPI_COMMANDS_DICT[cmd_tag][cmd_type][instr]

        # Check if the command is available for the instrument in use
        if command == "NOT_AVAILABLE":
            # do not send any request
            log.internal_warning(
                f"Command {cmd_type}:{cmd_tag} aborted as not available for instrument {self.instrument}"
            )
            return "COMMAND_NOT_AVAILABLE", validation

        if cmd_validation is not None:
            validation = (
                SCPI_COMMANDS_DICT[cmd_tag]["query"][instr],
                cmd_validation,
            )

        return command, validation

    def get_identification(self):
        """Get the identification information of an instrument.

        :return: the instrument's identification information
        """
        return self._send_scpi_command(
            cmd_tag="IDENTIFICATION",
            cmd_type="query",
        )

    def get_status_byte(self):
        """Get the status byte of an instrument.

        :return: the instrument's status byte
        """
        return self._send_scpi_command(
            cmd_tag="STATUS_BYTE",
            cmd_type="query",
        )

    def get_all_errors(self):
        """Get all errors of an instrument.

        return: list of off errors
        """
        return self._send_scpi_command(
            cmd_tag="ALL_ERRORS",
            cmd_type="query",
        )

    def reset(self):
        """Reset an instrument.

        :return: NO_VALIDATION status code
        """
        return self.visa_object.write("*RST")

    def self_test(self):
        """Performs a self-test of an instrument.

        :return: the query's response message
        """
        return self._send_scpi_command(cmd_tag="SELF_TEST", cmd_type="query")

    def get_remote_control_state(self):
        """Get the remote control mode (ON or OFF) of an instrument.

        :return: the remote control state
        """
        return self._send_scpi_command(
            cmd_tag="REMOTE_CONTROL",
            cmd_type="query",
        )

    def set_remote_control_on(self):
        """Enables the remote control of an instrument. The
        instrument will respond to all SCPI commands.

        :return: the writing operation's status code
        """
        return self._send_scpi_command(
            cmd_tag="REMOTE_CONTROL",
            cmd_type="write",
            cmd_payload="ON",
            cmd_validation=["REMOTE", "ON", "1"],
        )

    def set_remote_control_off(self):
        """Disable the remote control of an instrument. The
        instrument will respond to query and read commands only.

        :return: the writing operation's status code
        """
        return self._send_scpi_command(
            cmd_tag="REMOTE_CONTROL",
            cmd_type="write",
            cmd_payload="OFF",
            cmd_validation=["LOCAL", "OFF", "0", "NONE"],
        )

    def get_output_channel(self) -> str:
        """Get the currently selected output channel of an instrument.

        :return: the currently selected output channel
        """
        return self._send_scpi_command(
            cmd_tag="OUTPUT_CHANNEL",
            cmd_type="query",
        )

    def set_output_channel(self, channel: int) -> str:
        """Set the output channel of an instrument.

        :param channel: the output channel to select on the instrument

        :return: the writing operation's status code
        """
        return self._send_scpi_command(
            cmd_tag="OUTPUT_CHANNEL",
            cmd_type="write",
            cmd_payload=channel,
            cmd_validation=f"{channel}",
        )

    def get_output_state(self) -> str:
        """Get the output status (ON or OFF, enabled or disabled) of the
        currently selected channel of an instrument.

        :return: the output state (ON or OFF)
        """
        return self._send_scpi_command(
            cmd_tag="OUTPUT_ENABLE",
            cmd_type="query",
        )

    def enable_output(self) -> str:
        """Enable output on the currently selected output channel
        of an instrument.

        :return: the writing operation's status code
        """
        return self._send_scpi_command(
            cmd_tag="OUTPUT_ENABLE",
            cmd_type="write",
            cmd_payload="ON",
            cmd_validation=["ENABLE", "ENABLED", "ON", "1"],
        )

    def disable_output(self) -> str:
        """Disable output on the currently selected output channel
        of an instrument.

        :return: the writing operation's status code
        """
        return self._send_scpi_command(
            cmd_tag="OUTPUT_ENABLE",
            cmd_type="write",
            cmd_payload="OFF",
            cmd_validation=["DISABLE", "DISABLED", "OFF", "0"],
        )

    def get_nominal_voltage(self) -> str:
        """Query the nominal voltage of an instrument on the selected
        channel (in V).

        :return: the nominal voltage
        """
        return self._send_scpi_command(
            cmd_tag="NOMINAL_VOLTAGE",
            cmd_type="query",
        )

    def get_nominal_current(self) -> str:
        """Query the nominal current of an instrument on the selected
        channel (in A).

        :return: the nominal current
        """
        return self._send_scpi_command(
            cmd_tag="NOMINAL_CURRENT",
            cmd_type="query",
        )

    def get_nominal_power(self) -> str:
        """Query the nominal power of an instrument on the selected
        channel (in W).

        :return: the nominal power
        """
        return self._send_scpi_command(
            cmd_tag="NOMINAL_POWER",
            cmd_type="query",
        )

    def measure_voltage(self) -> str:
        """Return the measured output voltage of an instrument (in V).

        :return: the measured voltage
        """
        return self._send_scpi_command(
            cmd_tag="MEASURE_VOLTAGE",
            cmd_type="query",
        )

    def measure_current(self) -> str:
        """Return the measured output current of an instrument (in A).

        :return: the measured current
        """
        return self._send_scpi_command(
            cmd_tag="MEASURE_CURRENT",
            cmd_type="query",
        )

    def measure_power(self) -> str:
        """Return the measured output power of an instrument (in W).

        :return: the measured power
        """
        return self._send_scpi_command(
            cmd_tag="MEASURE_POWER",
            cmd_type="query",
        )

    def get_target_voltage(self) -> str:
        """Get the desired output voltage (in V) of an instrument.

        :return: the target voltage
        """
        return self._send_scpi_command(
            cmd_tag="VOLTAGE",
            cmd_type="query",
        )

    def get_target_current(self) -> str:
        """Get the desired output current (in A) of an instrument.

        :return: the target current
        """
        return self._send_scpi_command(
            cmd_tag="CURRENT",
            cmd_type="query",
        )

    def get_target_power(self) -> str:
        """Get the desired output power (in W) of an instrument.

        :return: the target power
        """
        return self._send_scpi_command(
            cmd_tag="POWER",
            cmd_type="query",
        )

    def set_target_voltage(self, value: float) -> str:
        """Set the desired output voltage (in V) of an instrument.

        :param value: value to be set on the instrument

        :return: the writing operation's status code
        """
        return self._send_scpi_command(
            cmd_tag="VOLTAGE",
            cmd_type="write",
            cmd_payload=value,
            cmd_validation=f"VALUE{{{value}}}",
        )

    def set_target_current(self, value: float) -> str:
        """Set the desired output current (in A) of an instrument.

        :param value: value to be set on the instrument

        :return: the writing operation's status code
        """
        return self._send_scpi_command(
            cmd_tag="CURRENT",
            cmd_type="write",
            cmd_payload=value,
            cmd_validation=f"VALUE{{{value}}}",
        )

    def set_target_power(self, value: float) -> str:
        """Set the desired output power (in W) of an instrument.

        :param value: value to be set on the instrument

        :return: the writing operation's status code
        """
        return self._send_scpi_command(
            cmd_tag="POWER",
            cmd_type="write",
            cmd_payload=value,
            cmd_validation=f"VALUE{{{value}}}",
        )

    def get_voltage_limit_low(self) -> str:
        """Returns the voltage lower limit (in V) of an instrument.

        :return: the query's response message
        """
        return self._send_scpi_command(
            cmd_tag="VOLTAGE_LIMIT_LOW",
            cmd_type="query",
        )

    def get_voltage_limit_high(self) -> str:
        """Returns the voltage upper limit (in V) of an instrument.

        :return: the query's response message
        """
        return self._send_scpi_command(
            cmd_tag="VOLTAGE_LIMIT_HIGH",
            cmd_type="query",
        )

    def get_current_limit_low(self) -> str:
        """Returns the current lower limit (in V) of an instrument.

        :return: the query's response message
        """
        return self._send_scpi_command(
            cmd_tag="CURRENT_LIMIT_LOW",
            cmd_type="query",
        )

    def get_current_limit_high(self) -> str:
        """Returns the current upper limit (in V) of an instrument.

        :return: the query's response message
        """
        return self._send_scpi_command(
            cmd_tag="CURRENT_LIMIT_HIGH",
            cmd_type="query",
        )

    def get_power_limit_high(self) -> str:
        """Returns the power upper limit (in W) of an instrument.

        :return: the query's response message
        """
        return self._send_scpi_command(
            cmd_tag="POWER_LIMIT_HIGH",
            cmd_type="query",
        )

    def set_voltage_limit_low(self, limit_value: float) -> str:
        """Set the voltage lower limit (in V) of an instrument.

        :param limit_value: limit value to be set on the instrument

        :return: the writing operation's status code
        """
        return self._send_scpi_command(
            cmd_tag="VOLTAGE_LIMIT_LOW",
            cmd_type="write",
            cmd_payload=limit_value,
            cmd_validation=f"VALUE{{{limit_value}}}",
        )

    def set_voltage_limit_high(self, limit_value: float) -> str:
        """Set the voltage upper limit (in V) of an instrument.

        :param limit_value: limit value to be set on the instrument

        :return: the writing operation's status code
        """
        return self._send_scpi_command(
            cmd_tag="VOLTAGE_LIMIT_HIGH",
            cmd_type="write",
            cmd_payload=limit_value,
            cmd_validation=f"VALUE{{{limit_value}}}",
        )

    def set_current_limit_low(self, limit_value: float) -> str:
        """Set the current lower limit (in A) of an instrument.

        :param limit_value: limit value to be set on the instrument

        :return: the writing operation's status code
        """
        return self._send_scpi_command(
            cmd_tag="CURRENT_LIMIT_LOW",
            cmd_type="write",
            cmd_payload=limit_value,
            cmd_validation=f"VALUE{{{limit_value}}}",
        )

    def set_current_limit_high(self, limit_value: float) -> str:
        """Set the current upper limit (in A) of an instrument.

        :param limit_value: limit value to be set on the instrument

        :return: the writing operation's status code
        """
        return self._send_scpi_command(
            cmd_tag="CURRENT_LIMIT_HIGH",
            cmd_type="write",
            cmd_payload=limit_value,
            cmd_validation=f"VALUE{{{limit_value}}}",
        )

    def set_power_limit_high(self, limit_value: float) -> str:
        """Set the power upper limit (in W) of an instrument.

        :param limit_value: limit value to be set on the instrument

        :return: the writing operation's status code
        """
        return self._send_scpi_command(
            cmd_tag="POWER_LIMIT_HIGH",
            cmd_type="write",
            cmd_payload=limit_value,
            cmd_validation=f"VALUE{{{limit_value}}}",
        )



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
JLink Flasher
*************

:module: flash_jlink

:synopsis: a Flasher adapter of the pylink-square library

.. currentmodule:: flash_jlink


"""

import logging
import pathlib
import typing

try:
    import pylink
except ImportError as e:
    raise ImportError(f"{e.name} dependency missing, consider installing pykiso with 'pip install pykiso[debugger]'")


from pykiso import Flasher

PathType = typing.Union[str, pathlib.Path]

log = logging.getLogger(__name__)


class JLinkFlasher(Flasher):
    """A Flasher adapter of the pylink-square library."""

    def __init__(
        self,
        binary: PathType = None,
        lib: PathType = None,
        serial_number: int = None,
        chip_name: str = "STM32L562QE",
        speed: int = 9600,
        verbose: bool = False,
        power_on: bool = False,
        start_addr: int = 0,
        xml_path: str = None,
        **kwargs,
    ):
        """Constructor.

        :param binary: path to the binary firmware file
        :param lib: path to the location of the JLink.so/JLink.DLL,
            usually automatically determined
        :param serial_number: optional debugger's S/N (required if
            many connected) (see pylink-square documentation)
        :param chip_name: see pylink-square documentation
        :param speed: see pylink-square documentation
        :param verbose: see pylink-square documentation
        :param power_on: see pylink-square documentation
        :param start_addr: see pylink-square documentation
        :param xml_path: device configuration (see pylink-square
            documentation)
        """
        self.lib = lib
        self.serial_number = serial_number if isinstance(serial_number, int) else None
        self.chip_name = chip_name
        self.speed = speed
        self.verbose = verbose
        self.power_on = power_on
        self.start_addr = start_addr
        self.start_addr = start_addr
        self.xml_path = xml_path
        self.jlink = None
        super().__init__(binary=binary, **kwargs)

    def open(self) -> None:
        """Initialize the flasher."""
        lib = self.lib
        if lib is not None:
            lib = pathlib.Path(self.lib).resolve()
            if lib.exists():
                lib = pylink.Library(dllpath=lib)
            else:
                log.internal_warning(f"No library file found at {lib}, using system default")
                lib = None

        # Define SEGGER J-Link interface
        self.jlink = pylink.JLink(lib=lib)

        # Open a connection to your J-Link
        self.jlink.open(serial_no=self.serial_number)

        # Define the the JLinkInterface
        self.jlink.set_tif(pylink.enums.JLinkInterfaces.JTAG)

        # Set Device config if file provided
        if self.xml_path is not None:
            self.jlink.exec_command(f"JLinkDevicesXMLPath {self.xml_path}")

        # Connect to the target
        self.jlink.connect(self.chip_name, self.speed, self.verbose)

        self.jlink.halt()
        self.jlink.reset()

        log.internal_debug("JLink connected to device")

    def close(self) -> None:
        """Close flasher and free resources."""
        self.jlink.close()

    def flash(self) -> None:
        """¨Perform firmware delivery.

        :raises pylink.JLinkException: if any hardware related error occurred
            during flashing.
        """
        log.internal_debug("flashing device")
        self.jlink.flash_file(str(self.binary), addr=self.start_addr, power_on=self.power_on)
        self.jlink.reset()
        log.internal_debug("flashing device successful")



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
CommunicationAuxiliary
**********************

:module: communication_auxiliary

:synopsis: Auxiliary used to send raw bytes via a connector instead of
    pykiso.Messages

.. currentmodule:: communication_auxiliary


"""
from __future__ import annotations

import functools
import logging
import queue
import threading
from contextlib import ContextDecorator
from typing import Any, Optional, Tuple

from pykiso import CChannel, Message
from pykiso.auxiliary import AuxiliaryInterface, close_connector, open_connector

log = logging.getLogger(__name__)


class _collect_messages(ContextDecorator):
    """Context manager and decorator for the communication auxiliary
    allowing messages collection (putting them in the queue).
    """

    def __init__(self, com_aux: CommunicationAuxiliary):
        """Constructor used to inherit some of communication auxiliary
        features.
        """
        self.com_aux = com_aux

    def __enter__(self):
        """Set the queue event to allow messages collection."""
        log.internal_debug("Start queueing received messages.")
        self.com_aux.queueing_event.set()

    def __exit__(self, *exc):
        """Clear queue event to stop messages collection."""
        log.internal_debug("Stop queueing received messages.")
        self.com_aux.queueing_event.clear()


class CommunicationAuxiliary(AuxiliaryInterface):
    """Auxiliary used to send raw bytes via a connector instead of pykiso.Messages."""

    def __init__(self, com: CChannel, **kwargs: dict) -> None:
        """Constructor.

        :param com: CChannel that supports raw communication
        """
        super().__init__(is_proxy_capable=True, tx_task_on=True, rx_task_on=True, **kwargs)
        self.channel = com
        self.queue_tx = queue.Queue()
        self.queueing_event = threading.Event()
        self.collect_messages = functools.partial(_collect_messages, com_aux=self)

    @open_connector
    def _create_auxiliary_instance(self) -> bool:
        """Open the connector communication.

        :return: True if the channel is correctly opened otherwise False
        """
        log.internal_info("Auxiliary instance created")
        return True

    @close_connector
    def _delete_auxiliary_instance(self) -> bool:
        """Close the connector communication.

        :return: always True
        """
        log.internal_info("Auxiliary instance deleted")
        return True

    def send_message(self, raw_msg: bytes, **kwargs) -> bool:
        """Send a raw message (bytes) via the communication channel.

        :param raw_msg: message to send
        :param kwargs: additional arguments to be passed to the
            underlying connector

        :return: True if command was executed otherwise False
        """
        return self.run_command("send", {"msg": raw_msg, **kwargs})

    def run_command(
        self,
        cmd_message: Any,
        cmd_data: Any = None,
        blocking: bool = True,
        timeout_in_s: int = None,
    ) -> bool:
        """Send a request by transmitting it through queue_in and
        populate queue_tx with the command verdict (successful or not).

        :param cmd_message: command to send
        :param cmd_data: data you would like to populate the command
            with
        :param blocking: If you want the command request to be
            blocking or not
        :param timeout_in_s: Number of time (in s) you want to wait
            for an answer

        :return: True if the request is correctly executed otherwise
            False
        """
        with self.lock:
            log.internal_debug(f"sending command '{cmd_message}' with payload {cmd_data} using {self.name} aux.")
            state = None
            self.queue_in.put((cmd_message, cmd_data))
            try:
                state = self.queue_tx.get(blocking, timeout_in_s)
                log.internal_debug(f"command '{cmd_message}' successfully sent for {self.name} aux")
            except queue.Empty:
                log.error(f"no feedback received regarding request {cmd_message} for {self.name} aux.")
        return state

    def receive_message(
        self,
        blocking: bool = True,
        timeout_in_s: float = None,
        receive_timestamp: bool = False,
    ) -> Optional[bytes | Tuple[bytes, int] | Tuple[bytes, int, float]]:
        """Receive a raw message.

        :param blocking: wait for message till timeout elapses?
        :param timeout_in_s: maximum time in second to wait for a response
        :param receive_timestamp: True if timestamp should be returned, False otherwise

        :returns: raw message
        """

        # Evaluate if we are in the context manager or not
        in_ctx_manager = False
        if self.queueing_event.is_set():
            in_ctx_manager = True

        log.internal_debug(f"retrieving message in {self} (blocking={blocking}, timeout={timeout_in_s})")
        # In case we are not in the context manager, we have a enable the receiver thread (and afterwards disable it)
        if not in_ctx_manager:
            self.queueing_event.set()
        response = self.wait_for_queue_out(blocking=blocking, timeout_in_s=timeout_in_s)
        if not in_ctx_manager:
            self.queueing_event.clear()

        log.internal_debug(f"retrieved message '{response}' in {self}")

        # if queue.Empty exception is raised None is returned so just
        # directly return it
        if response is None:
            return None

        msg = response.get("msg")
        remote_id = response.get("remote_id")
        timestamp = response.get("timestamp")

        # stay with the old return type to not making a breaking change
        if receive_timestamp:
            return (msg, remote_id, timestamp)
        elif remote_id and not receive_timestamp:
            return (msg, remote_id)
        return msg

    def clear_buffer(self) -> None:
        """Clear buffer from old stacked objects"""
        log.internal_info("Clearing buffer. Previous responses will be deleted.")
        with self.queue_out.mutex:
            self.queue_out.queue.clear()

    def _run_command(self, cmd_message: str, cmd_data: bytes = None) -> bool:
        """Run the corresponding command.

        :param cmd_message: command type
        :param cmd_data: payload data to send over CChannel

        :return: True if command is executed otherwise False
        """
        state = False
        if cmd_message == "send":
            try:
                self.channel.cc_send(**cmd_data)
                state = True
            except Exception:
                log.exception(f"encountered error while sending message '{cmd_data}' to {self.channel}")
        elif isinstance(cmd_message, Message):
            log.internal_debug(f"ignored command '{cmd_message} in {self}'")
        else:
            log.internal_warning(f"received unknown command '{cmd_message} in {self}'")

        self.queue_tx.put(state)

    def _receive_message(self, timeout_in_s: float) -> None:
        """Get a message from the associated channel. And put the message in
        the queue, if threading event is set.

        :param timeout_in_s: maximum amount of time (seconds) to wait
            for a message
        """
        try:
            rcv_data = self.channel.cc_receive(timeout=timeout_in_s)
            log.internal_debug(f"received message '{rcv_data}' from {self.channel}")
            msg = rcv_data.get("msg")
            if msg is not None and self.queueing_event.is_set():
                self.queue_out.put(rcv_data)
        except Exception:
            log.exception(f"encountered error while receiving message via {self.channel}")



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Acroname Control Auxiliary
**************************

:module: acroname_auxiliary

:synopsis: Auxiliary used to control acroname usb hubs.

.. currentmodule:: acroname_auxiliary

"""
import logging
from typing import Any, Optional

import brainstem
from brainstem.result import Result

from pykiso.auxiliary import AuxiliaryInterface
from pykiso.types import MsgType

log = logging.getLogger(__name__)

ERROR_MESSAGES = {
    0: "No Error occurred.",
    1: "Memory allocation/de-allocation error.",
    2: "Invalid parameters given.",
    3: "Entity, module or information not found.",
    4: "File name is to long.",
    5: "Module or resource is currently busy.",
    6: "An Input/Output error occurred.",
    7: "Invalid Mode or mode not accessible for current state.",
    8: "Write error occurred.",
    9: "Read error occurred.",
    10: "Unexpected end of file encountered.",
    11: "Resource not ready.",
    12: "Insufficient permissions.",
    13: "Request is outside of valid range.",
    14: "Size is incorrect for resource.",
    15: "Buffer was overrun or will be.",
    16: "Unable to parse command.",
    17: "Configuration is invalid.",
    18: "Timeout occurred.",
    19: "Could not initialize resource.",
    20: "Version mismatch",
    21: "Functionality unavailable or unimplemented.",
    22: "Duplicate request received.",
    23: "Request was canceled",
    24: "packet was invalid or had invalid contents.",
    25: "connection is no longer valid, or was closed.",
    26: "Requested entity does not exist.",
    27: "Command to short, not enough data to parse.",
    28: "Entity is not available, or does not exist.",
    29: "Option for given entity is invalid.",
    30: "Error allocating or acquiring a resource.",
    31: "Media not found or not available.",
    32: "Unknown error encountered.",
}


class AcronameAuxiliary(AuxiliaryInterface):
    """Auxiliary used to control acroname usb hubs"""

    MICROVOLT_TO_UNIT = {"uV": 1, "mV": 1e-3, "V": 1e-6}
    MICROAMP_TO_UNIT = {"uA": 1, "mA": 1e-3, "A": 1e-6}

    def __init__(self, serial_number: str | int = 0, **kwargs):
        """Constructor

        :param serial_number: serial number to connect to as hex string. Example "0x66F4859B"
        """
        super().__init__(
            is_proxy_capable=False,
            tx_task_on=False,
            rx_task_on=False,
            connector_required=False,
            **kwargs,
        )

        self.serial_number = int(serial_number, 16) if isinstance(serial_number, str) else serial_number
        self.stem = brainstem.stem.USBHub2x4()

    def _create_auxiliary_instance(self) -> bool:
        """Open the connector. Stop auxiliary if not possible.

        :return: True if successful
        """

        log.internal_info("Create auxiliary instance")

        result = self.stem.discoverAndConnect(brainstem.link.Spec.USB, self.serial_number)
        if result == (Result.NO_ERROR):
            result = self.stem.system.getSerialNumber()
            log.internal_info("Connected to USBStem with serial number: 0x%08X" % result.value)
        else:
            log.error("Could not connect to usb hub acroname")
            self.eval_result(result)
            return False

        return True

    def _delete_auxiliary_instance(self) -> bool:
        """Close the connector.

        :return: always True
        """
        log.internal_info("Delete auxiliary instance")
        try:
            self.stem.disconnect()
        except Exception:
            log.exception("Unable to close usb hub acroname.")
        return True

    @staticmethod
    def eval_result(result: Result) -> None:
        """Log error message if exist from acroname Result object.
        :param result: result code to evaluate
        """
        if result != (Result.NO_ERROR):
            log.error(f"Error occurred: {ERROR_MESSAGES[result]}")

    def set_port_enable(self, port: int) -> int:
        """Enable power and data lines for a USB port.

        :param port: the USB port number
        :return: brainstem error code. 0 if no error.
        """
        result = self.stem.usb.setPortEnable(port)
        self.eval_result(result)
        return result

    def set_port_disable(self, port: int) -> int:
        """Disable power and data lines for a USB port.

        :param port: the USB port number
        :return: brainstem error code. 0 if no error.
        """
        result = self.stem.usb.setPortDisable(port)
        self.eval_result(result)
        return result

    def get_port_current(self, port: int, unit: str = "A") -> Optional[float]:
        """Get the current through the power line for selected usb port.

        :param port: the USB port number
        :param unit: unit of the result in "uA", "mA" or "A". Default "A"
        :return: port current for given unit. None if unit is not supported.
        """
        port_current_ua = self.stem.usb.getPortCurrent(port)
        self.eval_result(port_current_ua.error)
        try:
            port_current = port_current_ua.value * self.MICROAMP_TO_UNIT[unit]

        except KeyError:
            log.error(f"Unit '{unit}' is not supported. Current value will be set to None.")
            return None

        return port_current

    def get_port_voltage(self, port: int, unit: str = "V") -> Optional[float]:
        """Get the voltage of the selected usb port.

        :param port: the USB port number
        :param unit: unit of the result in "uV", "mV" or "V". Default "V"
        :return: port voltage for given unit. None if unit is not supported.
        """
        port_voltage_uv = self.stem.usb.getPortVoltage(port)
        self.eval_result(port_voltage_uv.error)
        try:
            port_voltage = port_voltage_uv.value * self.MICROVOLT_TO_UNIT[unit]

        except KeyError:
            log.error(f"Unit '{unit}' is not supported. Voltage value will be set to None.")
            return None

        return port_voltage

    def get_port_current_limit(self, port: int, unit: str = "A") -> Optional[float]:
        """Get the current limit for the port.

        :param port: the USB port number
        :param unit: unit of the result in "uA", "mA" or "A". Default "A"
        :return: port current limit for given unit. None if unit is not supported.
        """
        result = self.stem.usb.getPortCurrentLimit(port)
        self.eval_result(result.error)
        try:
            port_current_limit = result.value * self.MICROAMP_TO_UNIT[unit]
        except KeyError:
            log.error(f"Unit '{unit}' is not supported. Port current value will be set to None.")
            return None

        return port_current_limit

    def set_port_current_limit(self, port: int, amps: float, unit: str = "A") -> int:
        """Set the current limit for the port. If the set limit is not achievable,
        devices will round down to the nearest available current limit setting.

        :param port: the USB port number
        :param amps: value for port current to set in "uA", "mA" or "A". Default "A"
        :param unit: unit for the value to set. Default Ampere
        :return: brainstem error code. 0 if no error.
        """
        try:
            micro_volt_value = int(amps / self.MICROAMP_TO_UNIT[unit])
        except KeyError:
            log.error(f"Unit '{unit}' is not supported. Port value won't be changed")
            return

        result = self.stem.usb.setPortCurrentLimit(port, micro_volt_value)
        self.eval_result(result)
        return result

    def _run_command(self, cmd_message: Any, cmd_data: Optional[bytes]) -> None:
        """Not used.

        Simply respect the interface.
        """

    def _receive_message(self, timeout_in_s: float) -> None:
        """Not used.

        Simply respect the interface.
        """



##########################################################################
# Copyright (c) 2010-2023 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Pytest command line extension
*****************************

:module: commandline

:synopsis: add pykiso-related options and ini values to pytest.

"""

from __future__ import annotations

from typing import TYPE_CHECKING

from .utils import export

if TYPE_CHECKING:
    from _pytest.config import Config
    from _pytest.config.argparsing import Parser
    from _pytest.main import PytestPluginManager


@export
def pytest_cmdline_main(config: Config):
    """Extract and reformat the CLI-provided tags as a dictionary.

    :param config: the pytest config instance.
    """

    def format_tags(unknown_args: list[str]) -> dict[str, list[str]]:
        formatted_args = dict()
        for arg in unknown_args:
            key, value = arg.split("=")
            key = key.replace("--", "", 1)
            key = key.replace("-", "_")
            formatted_args[key] = value.split(",")
        return formatted_args

    user_tags = config.getoption("tags", default=None)
    config.option.tags = format_tags(user_tags) if user_tags is not None else None


@export
def pytest_addoption(parser: Parser, pluginmanager: PytestPluginManager) -> None:
    """Add the 'tags' option to pytest's command line interface
    and the 'auxiliary_scope' ini value.

    :param parser: the command line parser.
    :param pluginmanager: not used.
    """
    group = parser.getgroup("pykiso")
    group.addoption(
        "--tags",
        nargs="+",
        dest="tags",
        help="Additional test tags to select tests to run depending on the tag value, "
        "e.g. '--tags variant=var1,var2 level=daily'",
    )
    parser.addini(
        "auxiliary_scope",
        default="session",
        type="string",
        help="Change this value to any of 'session', 'module', 'class', 'function' to change the "
        "scope for all auxiliaries. "
        "Note that if you define a scope of e.g. 'function', you won't be able to request a higher "
        "scope for an auxiliary by overriding the associated fixture.",
    )



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Integration Test Framework
**************************

:module: logging

:synopsis: Handles initialization of the loggers and custom logging levels.

.. currentmodule:: logging

"""
import importlib
import logging
import re
import sys
import time
from ast import literal_eval
from copy import deepcopy
from functools import partialmethod, wraps
from pathlib import Path
from typing import List, NamedTuple, Optional, Union

from .test_setup.dynamic_loader import PACKAGE
from .types import PathType

LEVELS = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
}


class LogOptions(NamedTuple):
    """
    Namedtuple containing the available options for logging configuration.
    """

    log_path: Optional[PathType]
    log_level: str
    report_type: str
    verbose: bool


# used to store the selected logging options
log_options: Optional[LogOptions] = None

# used to store the loggers that shouldn't be silenced
active_loggers = set()


def get_logging_options() -> LogOptions:
    """Simply return the previous logging options.

    :return: logging options log path, log level and report type
    """
    return log_options


def get_internal_level(log_level: Union[str, int]) -> int:
    """Retrieve the internal log level corresponding to the provided one.

    :param log_level: logging's log level for which the corresponding
        internal one should be retrieved. Can be a string (e.g. "DEBUG")
        or an integer (e.g. 10 or ``logging.DEBUG``)
    :return: the corresponding internal log level as an integer.
    """
    if isinstance(log_level, str):
        log_level = LEVELS[log_level]
    return log_level - 1


def add_internal_log_levels() -> None:
    """Create pykiso's internal log levels if not already done."""
    if not hasattr(logging, "INTERNAL_WARNING"):
        add_logging_level("INTERNAL_WARNING", get_internal_level(logging.WARNING))
        add_logging_level("INTERNAL_INFO", get_internal_level(logging.INFO))
        add_logging_level("INTERNAL_DEBUG", get_internal_level(logging.DEBUG))


def initialize_logging(
    log_path: Optional[PathType],
    log_level: str,
    verbose: bool,
    report_type: str = None,
    yaml_name: str = None,
) -> logging.Logger:
    """Initialize the logging.

    Sets the general log level, output file or STDOUT and the
    logging format.

    :param log_path: path to the logfile
    :param log_level: any of DEBUG, INFO, WARNING, ERROR
    :param verbose: activate internal kiso logging if True
    :param report_type: expected report type (junit, text,...)
    :param yaml_name: name of current yaml config file

    :returns: configured Logger
    """
    root_logger = logging.getLogger()
    # reset all previously added handlers to allow multiple calls
    root_logger.handlers = []

    log_format = logging.Formatter("%(asctime)s [%(levelname)s] %(module)s:%(lineno)d: %(message)s")
    # add internal kiso log levels
    add_internal_log_levels()

    # if log_path is given create a file handler
    if log_path is not None:
        log_path = Path(log_path)
        if log_path.suffix == "":
            log_path.mkdir(parents=True, exist_ok=True)
            fname = time.strftime(f"%Y-%m-%d_%H-%M-{yaml_name}.log")
            log_path = log_path / fname
        file_handler = logging.FileHandler(log_path, "a+")
        file_handler.setFormatter(log_format)
        # always include internal logs in log files
        file_handler.setLevel(get_internal_level(log_level))
        root_logger.addHandler(file_handler)

    # update logging options after having modified the log path
    global log_options
    log_options = LogOptions(log_path, log_level, report_type, verbose)

    # for junit use sys.stdout as stream, otherwise print to stderr
    stream = sys.stdout if report_type == "junit" else sys.stderr

    # for all report types add a StreamHandler
    stream_handler = logging.StreamHandler(stream)
    stream_handler.setFormatter(log_format)
    # disable internal logs if no verbose is wanted
    stream_handler.setLevel(LEVELS[log_level] if not verbose else get_internal_level(log_level))
    root_logger.addHandler(stream_handler)

    # set the root logger's level to the internal one if any of the provided options activates internal logging
    if verbose or log_path is not None:
        root_level = get_internal_level(log_level)
    else:
        root_level = log_level

    root_logger.setLevel(root_level)

    return logging.getLogger(__name__)


def add_logging_level(level_name: str, level_num: int):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `level_name` becomes an attribute of the `logging` module with the value
    `level_num`.
    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present
    Inspired by: https://stackoverflow.com/a/35804945

    Example
    -------
    >>> add_logging_level('KISO', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("KISO")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.kiso('so did this')
    >>> logging.KISO
    5

    :param level_name: name of the new level
    :param level_num: value of the new level
    """
    method_name = level_name.lower()

    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            # Use stacklevel=2 to skip this wrapper frame and show the actual caller
            kwargs.setdefault("stacklevel", 2)
            self._log(level_num, message, args, **kwargs)

    def log_to_root(message, *args, **kwargs):
        # Use stacklevel=2 to skip this wrapper frame and show the actual caller
        kwargs.setdefault("stacklevel", 2)
        logging.log(level_num, message, *args, **kwargs)

    if not hasattr(logging, level_name):
        logging.addLevelName(level_num, level_name)
        setattr(logging, level_name, level_num)
        setattr(logging.getLoggerClass(), method_name, log_for_level)
        setattr(logging, method_name, log_to_root)


def initialize_loggers(loggers: Optional[List[str]]) -> None:
    """Deactivate all external loggers except the specified ones.

    Store the loggers that should remain active globally to keep them
    activated when this function is called multiple times.

    :param loggers: list of logger names to keep activated
    """
    global active_loggers
    if loggers is None:
        loggers = list()
    else:
        loggers = deepcopy(loggers)

    # keyword 'all' should keep all loggers to the configured level
    if "all" in loggers:
        logging.internal_warning("All loggers are activated, this could lead to performance issues.")
        active_loggers |= set(logging.root.manager.loggerDict.keys())
        return
    # keep package and auxiliary loggers, store all the others to deactivate them
    relevant_loggers = {
        name: logger
        for name, logger in logging.root.manager.loggerDict.items()
        if not (name.startswith(PACKAGE) or name.endswith("auxiliary")) and not isinstance(logger, logging.PlaceHolder)
    }
    # keep child loggers
    children = [
        logger
        for logger in relevant_loggers.keys()
        for parent in loggers
        if (logger.startswith(parent) or parent.startswith(logger))
    ]
    loggers += children

    # store previous loggers to keep active (union of previous and current loggers)
    active_loggers |= set(loggers)

    # set the loggers that are not part of active_loggers to level WARNING
    loggers_to_deactivate = set(relevant_loggers) - set(active_loggers)
    for logger_name in loggers_to_deactivate:
        logging.getLogger(logger_name).setLevel(logging.WARNING)

    # but keep the other ones to the configured level as they could have deactivated by a previous call
    for logger_name in active_loggers:
        logging.getLogger(logger_name).setLevel(
            get_internal_level(log_options.log_level) if log_options.verbose else log_options.log_level
        )


def import_object(path: str) -> Union[None, logging.Logger]:
    """return the object based on the path.
        For example : logging.Logger will return Logger

    :param path: path to the object

    :return: object based on the path
    """
    if path:
        components = path.split(".")
        mod = importlib.import_module(".".join(components[:-1]))
        obj = getattr(mod, components[-1])
        if not issubclass(obj, logging.Logger):
            raise TypeError(f"{obj} is not derived from logging.Logger.")
        return obj
    else:
        return None


def remove_handler_from_logger(func):
    """Decorator that will remove all handlers of a logger after
        executing a function.

    :param func: function to execute
    """

    def wrapper(self, *arg, **kwargs):
        func(self, *arg, **kwargs)
        for handler in self.handlers:
            self.removeHandler(handler)

    return wrapper


def change_logger_class(log_level: str, verbose: bool, logger: str):
    """Change the class of all the logger of pykiso.

    :param log_level: level of the log
    :param logger: str of the path to the logger class
    """
    # Get the argument to initialize the logger if needed
    kwargs_log = {}
    # Search if the str has the following pattern name.name(name_arg=arg) or name.name
    arg_match = re.match(r"([^(]*)\(([^\)]+)\)", logger)
    if arg_match:
        logger_class = arg_match.group(1)
        arg_logger = arg_match.group(2).split(",")
        kwargs_log = {arg.split("=")[0]: literal_eval(arg.split("=")[1]) for arg in arg_logger}
    else:
        logger_class = logger

    # Import the logger class
    logger_class = import_object(logger_class)
    # We modify the init function so that the logger only need the name to be initialized
    if kwargs_log:
        logger_class.__init__ = partialmethod(logger_class.__init__, **kwargs_log)
    # Change logging.root since test can use logging.info for example
    logging.root = logger_class(name="root", level=LEVELS[log_level])
    # Remove the handler from the new logger else the handler will be called twice with the handler from the root
    logger_class.__init__ = remove_handler_from_logger(logger_class.__init__)

    # Replace already existing logger with the new class and change the parent
    for name, module in sys.modules.items():
        if name.startswith("pykiso"):
            if getattr(module, "log", None):
                module.log = logger_class(name=module.log.name, level=LEVELS[log_level])
                module.log.parent = logging.root

    # Setup the future logger class as the new class for the manager
    logging.Logger.manager.root = logging.root
    logging.setLoggerClass(logger_class)


def disable_logging(level: int = logging.CRITICAL):
    """Decorator used to disable all logging calls of severity 'level' and below.
        This function can be used to speed up test such as speed up uds transmission.

    :param level: logging level to disable, will also disable level below
        default logging.CRITICAL
    """

    def decorator(func):
        @wraps(func)
        def func_wrapper(self) -> None:
            logging.disable(level)
            func(self)
            logging.disable(logging.NOTSET)

        return func_wrapper

    return decorator
